/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xatom.h>

#include "dat.h"
#include "fns.h"

IXPServer *ixps;
char *ds;
Display *dpy;
File *core_files[CORE_LAST];
int mon;
int cid;
Atom wm_protocols;
Atom wm_delete;
Atom motif_wm_hints;
Ring *pages;
XRectangle rect;
Window root;

static Client zero_client = { 0 };

static int get_decoration(Window w) {

    struct MWMHints *mh = 0;
    long res = property(dpy, w, motif_wm_hints, motif_wm_hints, 5,
                        (unsigned char **) &mh);
    int ret = 1;

    if ((res >= 0) && mh) {
        ret = !((mh->decor == 0) && (mh->flags & MWM_HINTS_DECORATIONS));
        free((void *)mh);
    }
    return ret;
}

Client *alloc_client(Frame *parent, Window w, XWindowAttributes *wa) {

    long msize;
    char buf[MAX_BUF];
    char buf2[MAX_BUF];
    char buf3[MAX_BUF];
    int decor = get_decoration(w);
    Client *c = (Client *)malloc(sizeof(Client));
    Frame *f;
    XClassHint ch;

    if (!pages) alloc_frame(0, &rect, 0, 1);
    if (!parent) parent = pages->item;
    *c = zero_client;
    c->win = w;
    c->rect.x = wa->x;
    c->rect.y = wa->y;
    c->border = wa->border_width;
    c->rect.width = wa->width + 2 * c->border;
    c->rect.height = wa->height + 2 * c->border;
    assert(parent);
    f = c->frame = alloc_frame(parent, &c->rect, decor, decor);
    f->client = c;
    buf3[0] = buf2[0] = '\0';
    sget_ixppath(f->files[F_PREFIX], buf3, MAX_BUF);
    win_prop(dpy, c->win, XA_WM_NAME, buf2, MAX_BUF);
    snprintf(buf, MAX_BUF, "%s/client/name", buf3);
    c->files[C_NAME] = screate_ixpfile(ixps, buf, buf2);
    if (XGetClassHint(dpy, c->win, &ch)) {
        snprintf(buf, MAX_BUF, "%s/client/class", buf3);
        c->files[C_CLASS] = screate_ixpfile(ixps, buf, ch.res_class);
        snprintf(buf, MAX_BUF, "%s/client/instance", buf3);
        c->files[C_INSTANCE] = screate_ixpfile(ixps, buf, ch.res_name);
    }
    else {
        snprintf(buf, MAX_BUF, "%s/client/class", buf3);
        c->files[C_CLASS] = ixp_create(ixps, buf);
        snprintf(buf, MAX_BUF, "%s/client/instance", buf3);
        c->files[C_INSTANCE] = ixp_create(ixps, buf);
    }
    c->reparenting++;
    set_win_state(c, WithdrawnState);
    XSetWindowBorderWidth(dpy, c->win, 0);
    if (f->decor)
        XReparentWindow(dpy, c->win, f->win,
                        int_val(f->files[F_BORDER_W]->content),
                        int_val(f->files[F_TAB_H]->content));
    XSelectInput(dpy, c->win, PropertyChangeMask);
    c->proto = win_proto(c->win);
    XGetTransientForHint(dpy, c->win, &c->trans);
    /* size hints */
    if (!XGetWMNormalHints(dpy, c->win, &c->size, &msize) || !c->size.flags)
        c->size.flags = PSize;
    return c;
}

void handle_client_property(Client *c, XPropertyEvent *e) {

    char buf[MAX_BUF];
    long msize;

    buf[0] = '\0';
    if (e->state == PropertyDelete) return; /* ignore */

    if (e->atom == wm_protocols) {
        /* update */
        c->proto = win_proto(c->win);
        return;
    }

    switch (e->atom) {
    case XA_WM_NAME:
        win_prop(dpy, c->win, XA_WM_NAME, buf, MAX_BUF);
        if (strlen(buf)) {
            if (c->files[C_NAME]->content) free(c->files[C_NAME]->content);
            c->files[C_NAME]->content = strdup(buf);
            c->files[C_NAME]->size = strlen(buf);
        }
        if (c->frame) draw_child(c->frame);
        break;
    case XA_WM_TRANSIENT_FOR:
        XGetTransientForHint(dpy, c->win, &c->trans);
        break;
    case XA_WM_NORMAL_HINTS:
        if (!XGetWMNormalHints(dpy, c->win, &c->size, &msize)
             || !c->size.flags)
        {
            c->size.flags = PSize;
        }
        break;
    }
}

void free_client(Client *c) {
    free_frame(c->frame);
    free(c);
}

void invoke_class_instance_hook(Client *c, char *hook) {
    char buf[MAX_BUF];
    File *f;

    if (!core_files[CORE_ACTION_LOOKUP]->content ||
        !c->files[C_CLASS]->content || !c->files[C_INSTANCE]->content)
        return;

    snprintf(buf, sizeof(buf), "/default/client/%s:%s/%s",
             c->files[C_CLASS]->content ? (char *)c->files[C_CLASS]->content : "",
             c->files[C_INSTANCE]->content ? (char *)c->files[C_INSTANCE]->content : "", hook);
    f = ixp_walk(ixps, buf);
    if (f && f->content)
        spawn_expanded(dpy, ds, core_files[CORE_ACTION_LOOKUP]->content,
                       f->content);
}

void configure_client(Client *c) {

    XConfigureEvent e;
    e.type = ConfigureNotify;
    e.event = c->win;
    e.window = c->win;
    e.x = c->rect.x;
    e.y = c->rect.y;
    e.width = c->rect.width;
    e.height = c->rect.height;
    e.border_width = c->border;
    e.above = None;
    e.override_redirect = False;
    XSendEvent(dpy, c->win, False, StructureNotifyMask, (XEvent*)&e);
}

void close_client(Client *c) {

    if (c->proto & PROTO_DEL) {
        send_message(dpy, c->win, wm_protocols, wm_delete);
    }
    else {
        XKillClient(dpy, c->win);
    }
}

void gravitate(Client *c, unsigned int tabh, unsigned int bw, int invert) {

    int dx = 0, dy = 0;
    int gravity = NorthWestGravity;

    if (c->size.flags & PWinGravity) {
        gravity = c->size.win_gravity;
    }

    /* y */
    switch (gravity) {
    case StaticGravity:
    case NorthWestGravity:
    case NorthGravity:
    case NorthEastGravity:
        dy = bw + tabh;
        break;
    case EastGravity:
    case CenterGravity:
    case WestGravity:
        dy = -(c->rect.height / 2) + bw + tabh;
        break;
    case SouthEastGravity:
    case SouthGravity:
    case SouthWestGravity:
        dy = -(c->rect.height + bw);
        break;
    default: /* don't care */
        break;
    }

    /* x */
    switch (gravity) {
    case StaticGravity:
    case NorthWestGravity:
    case WestGravity:
    case SouthWestGravity:
        dx = bw;
        break;
    case NorthGravity:
    case CenterGravity:
    case SouthGravity:
        dx = -(c->rect.width / 2) + bw;
        break;
    case NorthEastGravity:
    case EastGravity:
    case SouthEastGravity:
        dx = -(c->rect.width + bw);
        break;
    default: /* don't care */
        break;
    }

    if (invert) {
        dx = -dx;
        dy = -dy;
    }

    c->rect.x += dx;
    c->rect.y += dy;
}

