/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>

#include "dat.h"
#include "fns.h"

Display *dpy;
char *ds;
IXPServer *ixps;
int mon;
File *core_files[CORE_LAST];
Ring *detached;
Ring *pages;
XRectangle rect;
Window root;
Frame *zoomed;
Client *sel;

Atom wm_delete;
Atom wm_protocols;
Atom wm_state;

Cursor normal_cursor;
Cursor w_cursor;
Cursor e_cursor;
Cursor n_cursor;
Cursor s_cursor;
Cursor nw_cursor;
Cursor ne_cursor;
Cursor sw_cursor;
Cursor se_cursor;

static void _attach(void *obj, char *cmd);
static void _detach(void *obj, char *cmd);
static void _select(void *obj, char *cmd);
static void new_page(void *obj, char *cmd);
static void _close_client(void *obj, char *cmd);
static void quit(void *obj, char *cmd);
static void destroy_frame(void *obj, char *cmd);

/* action table for /ctl namespace */
Action core_acttbl[8] = {
    {"attach", _attach},
    {"detach", _detach},
    {"select", _select},
    {"new", new_page},
    {"destroy", destroy_frame},
    {"close", _close_client},
    {"exec", _exec},
    {"quit", quit}
};

int comp_obj(void *c1, void *c2) {
    return c1 == c2;
}

static Action *get_frame_acttbl(File *f, Ring *root, void **obj, int *num_act) {

    Ring *r;
    Action *res = 0;

    for (r = root; r; r = r->next) {
        Frame *fr = r->item;
        if (fr->files[F_CTL] == f) {
            *num_act = fr->num_act;
            *obj = fr;
            return fr->acttbl;
        }
        if (fr->sel)
            res = get_frame_acttbl(f, fr->sel, obj, num_act);
        if (res) return res;
        if (r->next == root) break;
    }
    return 0;
}

void run_action(File *f) {

    int num_act = 0;
    Action *acttbl = 0;
    void *obj = 0;

    if (f == core_files[CORE_CTL]) {
        num_act = 8;
        acttbl = core_acttbl;
    }
    else acttbl = get_frame_acttbl(f, pages, &obj, &num_act);

    if (zoomed && zoomed != obj) {
        zoomed->rect = zoom_rect;
        resize(zoomed);
        zoomed = 0;
    }
    if (acttbl) {
        int i;
        size_t len;
        for (i = 0; i < num_act; i++) {
            len = strlen(acttbl[i].name);
            if (!strncmp(acttbl[i].name, (char *)f->content, len)) {
                if (f->size > len) {
                    acttbl[i].func(obj, &((char *)f->content)[len + 1]);
                }
                else {
                    acttbl[i].func(obj, 0);
                }
                break;
            }
        }
    }
    else {
        fprintf(stderr, "wmii: unknown action '%s'\n"
                        "      or invalid ctl device\n", (char *)f->content);
    }
}

/* {{{ action implementations */

void _exec(void *obj, char *cmd) {
    spawn(dpy, ds, cmd);
}

static void quit(void *obj, char *cmd) {
    ixps->runlevel = SHUTDOWN;
}

void grab_client(Client *c) {
    XGrabButton(dpy, AnyButton, AnyModifier, c->win, False,
                ButtonPressMask, GrabModeAsync, GrabModeAsync,
                None, None);
}

void ungrab_client(Client *c) {
    XUngrabButton(dpy, AnyButton, AnyModifier, c->win);
}


void invoke_page_event(File *f) {
    if (!core_files[CORE_ACTION_LOOKUP]->content || !f->content) return;
    spawn_expanded(dpy, ds, core_files[CORE_ACTION_LOOKUP]->content,
                   f->content);
}

/* {{{ attach/detach stuff */

void attach(Frame *parent, Frame *child) {

    Layout *l;
    Client *c = child->client;

    if (c) invoke_class_instance_hook(c, "before-attach");

    /* transient stuff */
    if (c && c->trans) {
        Client *t = win_to_client(pages, c->trans);
        if (t && t->frame) focus(t->frame, 1);
        if (root_frame(parent) != pages->item)
            parent = pages->item;
        set_ignore_layout(child, "1");
    }
    l = match_layout(parent->layout, size_of_frame(parent)
                     + !int_val(child->files[F_IGNORE_LAYOUT]->content));

    if (sel) grab_client(sel);
    if (parent->parent) XMapRaised(dpy, parent->win);

    /* attach to layout */
    if (parent->parent) {
       if (!parent->client && l) {
           attach_frame_to_frame(parent, child);
           XMapRaised(dpy, child->win);
       }
       else {
           attach(parent->parent, child);
           return;
       }
    }
    else {
        attach_frame_to_frame(parent, child);
        if (parent->decor) XMapRaised(dpy, child->win);
    }

    /* final stuff */
    if (!int_val(child->files[F_IGNORE_LAYOUT]->content)) resize(parent);
    else resize(child);
    if (c) {
        sel = c;
        XMapRaised(dpy, c->win);
        XAddToSaveSet(dpy, c->win);
        set_win_state(c, NormalState);
        ungrab_client(c);
        XSetInputFocus(dpy, c->win, RevertToPointerRoot, CurrentTime);
    }

    parent->files[F_SELECTED_FRAME]->content = child->files[F_PREFIX]->content;
    core_files[CORE_FRAME_SEL]->content = child->files[F_PREFIX]->content;
    if (c) invoke_class_instance_hook(c, "after-attach");
    invoke_page_event(core_files[CORE_EVENT_PAGE_UPDATE]);
}

Frame *top_frame(Frame *child) {
    Frame *p;
    if (!child) return 0;

    p = root_frame(child);
    if (p->frames) return p->frames->item;
    return p;
}

void detach(Frame *f, int client_destroyed) {

    Client *c = f->client;
    Frame *fp = f->parent;
    Frame *sf = 0;
    if (!fp) return; /* page detaches aren't allowed */

    if (c && !client_destroyed) {
        invoke_class_instance_hook(c, "before-detach");
        set_win_state(c, WithdrawnState);
        XUnmapWindow(dpy, c->win);
    }
    if (f->decor) XUnmapWindow(dpy, f->win);

    detach_frame_from_frame(f);
    detached = r_attach(detached, f);

    if (fp && fp->files[F_SELECTED_FRAME]->content == f->files[F_PREFIX]->content)
        fp->files[F_SELECTED_FRAME]->content = 0;
    if (core_files[CORE_FRAME_SEL]->content == f->files[F_PREFIX]->content)
        core_files[CORE_FRAME_SEL]->content = 0;

    if (fp->parent && size_of_frame(fp) < 2)
        fp = apply_layout(fp);
    else resize(fp);

    /* finally focus next frame on the stack */
    sf = top_frame(fp ? fp : pages->item);
    if (sf) focus(sf, 1);
    if (c && !client_destroyed)
        invoke_class_instance_hook(c, "after-detach");
    invoke_page_event(core_files[CORE_EVENT_PAGE_UPDATE]);
}

/* }}} */

void focus(Frame *f, int raise) {

    Frame *fr, *p;

    /* grab old selected client */
    if (sel) grab_client(sel);

    p = root_frame(f);
    if (p != pages->item) {
        /* focus another page */
        hide_page(pages->item);
        pages = r_find(pages, p, comp_obj);
        show_page(pages->item);
        core_files[CORE_SELECTED_PAGE]->content = p->files[F_PREFIX]->content;
        invoke_page_event(core_files[CORE_EVENT_PAGE_SEL]);
    }

    if (raise && f->parent) {
        fr = f;
        while ((fr = fr->parent))
            if (fr->parent && fr->decor) XRaiseWindow(dpy, fr->win);
        if (f->decor) XRaiseWindow(dpy, f->win);
        else if (f->client) XRaiseWindow(dpy, f->client->win);
    }

    /* focussing on stack */
    for (fr = f; !fr->client && fr->sel; fr = fr->sel->item);
    if (fr->client) {
        sel = fr->client;
        ungrab_client(sel);
        XSetInputFocus(dpy, sel->win, RevertToPointerRoot, CurrentTime);
    }

    p->frames = r_top(p->frames, f, comp_obj);
    assert(p->frames);
    if (f->parent) {
        f->parent->sel = r_find(f->parent->childs, f, comp_obj);
        f->parent->files[F_SELECTED_FRAME]->content = f->files[F_PREFIX]->content;
    }
    core_files[CORE_FRAME_SEL]->content = f->files[F_PREFIX]->content;
}

static void _destroy_frame(Frame *f) {
    while (f->sel) _destroy_frame(f->sel->item);
    if (f->client) {
        if (f->client == sel) sel = 0;
        ungrab_client(f->client);
        XUnmapWindow(dpy, f->client->win);
        set_win_state(f->client, WithdrawnState);
        if (f->decor) XUnmapWindow(dpy, f->win);
        detach_frame_from_frame(f);
        detached = r_attach(detached, f);
    }
    else {
        if (f->parent) {
            XUnmapWindow(dpy, f->win);
            detach_frame_from_frame(f);
        }
        free_frame(f);
    }
}

static Frame *find_frame_by_name(Ring *root, char *name) {

    Ring *r;
    size_t len;
    if (!name || name[0] == '\0') return 0;
    len = strlen(name);
    for (r = root; r; r = r->next) {
        Frame *f = r->item;
        Frame *res;
        if (!strncmp(f->files[F_NAME]->content, name, len))
            return f;
        if (root != detached) {
            res = find_frame_by_name(f->childs, name);
            if (res) return res;
        }
        if (r->next == root) break;
    }
    return 0;
}

static void destroy_frame(void *obj, char *cmd) {

    Frame *f, *fp;
    if (!pages) return;
    if (cmd) f = find_frame_by_name(pages, cmd);
    else f = top_frame(pages->item);
    if (!f) return;
    fp = f->parent;
    _destroy_frame(f);
    if (fp) {
       if (size_of_frame(fp) < 2) apply_layout(fp);
       else resize(fp);
    }
    if (pages) {
        f = pages->item;
        show_page(f);
        core_files[CORE_SELECTED_PAGE]->content = f->files[F_PREFIX]->content;
        invoke_page_event(core_files[CORE_EVENT_PAGE_SEL]);
        focus(top_frame(f), 1);
    }
}

static void _attach(void *obj, char *cmd) {
    Frame *f;
    if (detached) {
        if (cmd) f = find_frame_by_name(detached, cmd);
        else f = detached->item;
        if (!f) return;
        detached = r_detach(detached, f);
        if (!pages) alloc_frame(0, &rect, 0, 1);
        if (f->decor) attach(top_frame(pages->item), f);
        else attach(pages->item, f);
    }
}

static void _detach(void *obj, char *cmd) {
    Frame *f;
    if (cmd && pages) f = find_frame_by_name(pages, cmd);
    else f = pages ? top_frame(pages->item) : 0;
    if (f && f->parent) detach(f, 0);
}

static void _select(void *obj, char *cmd) {
    Frame *f = 0;
    if (cmd && pages)
        f = find_frame_by_name(pages, cmd);
    if (f) {
        focus(f, 1);
        draw_frame(pages->item);
    }
}

static void new_page(void *obj, char *cmd) {
    Frame *p;
    size_t len;
    if (pages && pages->item) hide_page(pages->item);
    p = alloc_frame(0, &rect, 0, 1);
    if (cmd) {
        len = strlen(cmd);
        if (!len) return;
        free(p->files[F_NAME]->content);
        p->files[F_NAME]->content = strdup(cmd);
        p->files[F_NAME]->size = len;
    }
}

static void _close_client(void *obj, char *cmd) {
    if (sel) close_client(sel);
}

/* }}} */

/* {{{ core helpers */

Align cursor_to_align(Cursor cursor) {

    if (cursor == w_cursor) return WEST;
    else if (cursor == nw_cursor) return NWEST;
    else if (cursor == n_cursor) return NORTH;
    else if (cursor == ne_cursor) return NEAST;
    else if (cursor == e_cursor) return EAST;
    else if (cursor == se_cursor) return SEAST;
    else if (cursor == s_cursor) return SOUTH;
    else if (cursor == sw_cursor) return SWEST;
    return CENTER; /* should not happen */
}

Client *win_to_client(Ring *root, Window w) {

    Ring *r;
    Client *c = 0;

    for (r = root; r; r = r->next) {
        Frame *fr = r->item;
        if (fr->client && fr->client->win == w) return fr->client;
        if (fr->sel)
            c = win_to_client(fr->sel, w);
        if (c) return c;
        if (r->next == root) break;
    }
    return 0;
}

Frame *win_to_frame(Ring *root, Window w) {

    Ring *r;
    Frame *f = 0;

    for (r = root; r; r = r->next) {
        Frame *fr = r->item;
        if (fr->win == w)
            return fr;
        if (fr->sel)
            f = win_to_frame(fr->sel, w);
        if (f) return f;
        if (r->next == root) break;
    }
    return 0;
}

void scan_wins() {
    int i;
    unsigned int num;
    Window *wins;
    XWindowAttributes wa;
    Window d1, d2;
    Client *c;

    if (XQueryTree(dpy, root, &d1, &d2, &wins, &num)) {
        for (i = 0; i < num; i++) {
            XGetWindowAttributes(dpy, wins[i], &wa);
            if (wa.override_redirect
                || wa.map_state != IsViewable
                || XGetTransientForHint(dpy, wins[i], &d1))
                continue;
            c = alloc_client(0, wins[i], &wa);
            if (c->frame->decor) attach(top_frame(pages->item), c->frame);
            else attach(pages->item, c->frame);
        }
    }
    if (wins) XFree(wins);
}

void *get_func(void *acttbl[][2], int rows, char *fname) {

    int i;
    for (i = 0; i < rows; i++) {
       if (!strncmp((char *)acttbl[i][0], fname, MAX_BUF)) {
          return acttbl[i][1];
       }
    }
    return 0;
}

int win_proto(Window w) {
    Atom *protocols;
    long res;
    int protos = 0;
    int i;

    res = property(dpy, w, wm_protocols, XA_ATOM,
                   20L, ((unsigned char **)&protocols));
    if (res <= 0) {
        return protos;
    }

    for (i = 0; i < res; i++) {
        if (protocols[i] == wm_delete) {
            protos |= PROTO_DEL;
        }
    }
    free((char *) protocols);
    return protos;
}

int win_state(Window w) {

    /* state hints */
    XWMHints *hints = XGetWMHints(dpy, w);
    int res;

    long *prop = 0;
    if (property(dpy, w, wm_state, wm_state,
                2L, ((unsigned char**)&prop)) > 0)
    {
        res = (int)*prop;
        free((long *)prop);
    }
    else {
        res = hints ? hints->initial_state : NormalState;
    }

    if (hints) {
        free(hints);
    }

    return res;
}

void set_win_state(Client *c, int state) {

    long data[2];

    data[0] = (long) state;
    data[1] = (long) None;

    c->state = state;
    XChangeProperty(dpy, c->win, wm_state, wm_state, 32,
                    PropModeReplace, (unsigned char *)data, 2);
}

/* }}} */

/* {{{ IXP callbacks */

static void update_xorgc_color(Ring *root) {

    Ring *r;

    for (r = root; r; r = r->next) {
        Frame *f = r->item;
        XSetForeground(dpy, f->xorgc, xorcolor.pixel);
        if (f->sel) update_xorgc_color(f->sel);
        if (r->next == root) break;
    }
}

void handle_after_write(IXPServer *s, File *f) {

    if (!strncmp(f->name, "ctl", 4)) {
        if (f->content) run_action(f);
    }
    else if (f == core_files[CORE_TRANS_COLOR]) {
        unsigned long col[1];
        col[0] = xorcolor.pixel;
        XFreeColors(dpy, DefaultColormap(dpy, mon), col, 1, 0);
        XAllocNamedColor(dpy, DefaultColormap(dpy, mon),
                         core_files[CORE_TRANS_COLOR]->content,
                         &xorcolor, &xorcolor);
        update_xorgc_color(pages);
    }
    check_event(0);
}

/* }}} */

void invoke_rc(char *mode, void (*spawn_func)(void *, char *, char*)) {
    char rcfile[256];
    char cmd[256];

    /* first run $HOME/.wmii/rc if any */
    snprintf(rcfile, 256, "%s/.wmii/rc", getenv("HOME"));
    if (!access(rcfile, X_OK)) {
        snprintf(cmd, 256, "%s %s", rcfile, mode);
        spawn_func(dpy, ds, cmd);
    }
    else {
        snprintf(rcfile, 256, "%s/wmii/rc", CONFPREFIX);
        if (!access(rcfile, X_OK)) {
            snprintf(cmd, 256, "%s %s", rcfile, mode);
            spawn_func(dpy, ds, cmd);
        }
        else {
            fprintf(stderr, "wmii: warning, cannot execute '%s',"
                    " using internal failsafe fallback\n", rcfile);
        }
    }
}
