/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "dat.h"
#include "fns.h"

Display *dpy;
Ring *pages;
Ring *detached;

Atom wm_state;
Atom wm_change_state;
Atom wm_protocols;

Cursor normal_cursor;

/* local functions */
static void handle_buttonpress(XEvent *e);
static void handle_configurerequest(XEvent *e);
static void handle_destroynotify(XEvent *e);
static void handle_expose(XEvent *e);
static void handle_maprequest(XEvent *e);
static void handle_motionnotify(XEvent *e);
static void handle_propertynotify(XEvent *e);
static void handle_unmapnotify(XEvent *e);

void (*handler[LASTEvent]) (XEvent *);

void init_event_hander() {

    int i;
    /* init event handler */
    for (i = 0; i < LASTEvent; i++) {
        handler[i] = 0;
    }
    handler[ButtonPress]      = handle_buttonpress;
    handler[ConfigureRequest] = handle_configurerequest;
    handler[DestroyNotify]    = handle_destroynotify;
    handler[Expose]           = handle_expose;
    handler[MapRequest]       = handle_maprequest;
    handler[MotionNotify]     = handle_motionnotify;
    handler[PropertyNotify]   = handle_propertynotify;
    handler[UnmapNotify]      = handle_unmapnotify;
}

void check_event(Connection *c) {

    XEvent ev;
    while (XPending(dpy)) {
        XNextEvent(dpy, &ev);
        /* main evet loop */
        if (handler[ev.type]) {
            /* call handler */
            (handler[ev.type])(&ev);
        }
    }
}

static void handle_buttonpress(XEvent *e) {

    Client *c;
    Frame *f = win_to_frame(pages, e->xbutton.window);
    if (f) handle_frame_buttonpress(&e->xbutton, f);

    c = win_to_client(pages, e->xbutton.window);
    if (c) {
        focus(c->frame, 1);
        draw_frame(pages->item);
    }
}

static void handle_configurerequest(XEvent *e) {

    XConfigureRequestEvent *ev = &e->xconfigurerequest;
    XWindowChanges wc;
    Client *c;
    unsigned int bw = 0, tabh = 0;
    unsigned int has_frame;

    c = win_to_client(pages, ev->window);
    has_frame = c && c->frame && c->frame->decor;

    if (has_frame) {
        if (c->frame->parent->parent) {
            resize(c->frame);
            resize(c->frame);
            return;
        }
        bw = int_val(c->frame->files[F_BORDER_W]->content);
        tabh = int_val(c->frame->files[F_TAB_H]->content);
    }

    if (c) {

        gravitate(c, tabh, bw, 1);

        if (ev->value_mask & CWX) c->rect.x = ev->x;
        if (ev->value_mask & CWY) c->rect.y = ev->y;
        if (ev->value_mask & CWWidth) c->rect.width = ev->width;
        if (ev->value_mask & CWHeight) c->rect.height = ev->height;
        if (ev->value_mask & CWBorderWidth) c->border = ev->border_width;

        gravitate(c, tabh, bw, 0);

        if (has_frame) {
            c->frame->rect.x = wc.x = c->rect.x - bw;
            c->frame->rect.y = wc.y = c->rect.y - tabh - bw;
            c->frame->rect.width = wc.width = c->rect.width + 2 * bw;
            c->frame->rect.height = wc.height = c->rect.height + bw + tabh;
            wc.border_width = 1;
            wc.sibling = ev->above;
            wc.stack_mode = ev->detail;
            XConfigureWindow(dpy, c->frame->win, ev->value_mask, &wc);
            configure_client(c);
        }
    }

    wc.x = ev->x;
    wc.y = ev->y;

    if (has_frame) {
        /* if so, then bw and tabh are already initialized */
        wc.x = bw;
        wc.y = tabh;
    }

    wc.width = ev->width;
    wc.height = ev->height;
    wc.border_width = 0;
    wc.stack_mode = ev->detail;
    ev->value_mask |= CWBorderWidth;
    XConfigureWindow(dpy, ev->window, ev->value_mask, &wc);

}

static void handle_destroynotify(XEvent *e) {

    XDestroyWindowEvent *ev = &e->xdestroywindow;
    Client *c = win_to_client(pages, ev->window);
    if (!c) return; /* unhandled window */
    if (sel == c) sel = 0;
    detach(c->frame, 1);
    detached = r_detach(detached, c->frame);
    free_client(c); /* frees also the frame */
}

static void handle_expose(XEvent *e) {

    if (e->xexpose.count == 0) {
        Frame *f = win_to_frame(pages, e->xbutton.window);
        if (f) draw_frame(f);
        XSync(dpy, False);
    }
}

static void handle_maprequest(XEvent *e) {

    XMapRequestEvent *ev = &e->xmaprequest;
    XWindowAttributes wa;
    Client *c;

    XGetWindowAttributes(dpy, ev->window, &wa);
    if (wa.override_redirect) return;

    if (!(c = win_to_client(pages, ev->window)))
        c = alloc_client(0, ev->window, &wa);
    XSync(dpy, False);
    fprintf(stderr, "%s", "client allocated\n");

    switch (c->state) {
    case WithdrawnState:
    case IconicState:
        if (c->frame->decor) attach(top_frame(pages->item), c->frame);
        else attach(pages->item, c->frame);
    XSync(dpy, False);
    fprintf(stderr, "%s", "client attached\n");
        draw_frame(pages->item);
        break;
    case NormalState:
        if (c->frame)
            focus(c->frame, 1);
        break;
    }
}

static void handle_motionnotify(XEvent *e) {

    Frame *f = win_to_frame(pages, e->xmotion.window);
    Cursor cursor;
    if (f) {
        cursor = cursor_for_xy(f, e->xmotion.x, e->xmotion.y);
        if (cursor != f->cursor) {
            f->cursor = cursor;
            install_cursor(dpy, f->win, cursor);
        }
    }
}

static void handle_propertynotify(XEvent *e) {

    static Atom x_rootpixmap = None;
    XPropertyEvent *ev = &e->xproperty;
    Client *c = win_to_client(pages, ev->window);

    if (c) {
        handle_client_property(c, ev);
        return;
    }

    /* check root pixmap */

    if (x_rootpixmap == None) {
        x_rootpixmap = XInternAtom(dpy, "_XROOTPMAP_ID", True);
    }

    if ((x_rootpixmap != None) && (ev->atom == x_rootpixmap)) {
        if (ev->window == root && pages) {
            draw_frame(pages->item);
        }
    }
}

static void handle_unmapnotify(XEvent *e) {

    XUnmapEvent *ev = &e->xunmap;
    Client *c;

    c = win_to_client(pages, ev->window);
    if (c) {
        if (!c->reparenting) {
            detach(c->frame, 1);
            XReparentWindow(dpy, c->win, root, c->frame->rect.x + c->rect.x,
                            c->frame->rect.y + c->rect.y);
            detached = r_detach(detached, c->frame);
            free_client(c);
            if (pages) draw_frame(pages->item);
        }
        if (c->reparenting) c->reparenting--;
    }
}
