/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "dat.h"
#include "fns.h"

Display *dpy;
char *ds;
IXPServer *ixps;
File *defaults[F_LAST];
File *core_files[CORE_LAST];
Ring *detached;
Ring *pages, *first_page;
int mon;
int fid;
Frame *zoomed;
Window root;
XColor xorcolor;
XRectangle rect;

Cursor normal_cursor;
Cursor resize_cursor;
Cursor move_cursor;
Cursor drag_cursor;
Cursor w_cursor;
Cursor e_cursor;
Cursor n_cursor;
Cursor s_cursor;
Cursor nw_cursor;
Cursor ne_cursor;
Cursor sw_cursor;
Cursor se_cursor;

static Frame zero_frame = { 0 };

static void __resize(void *obj, char *cmd);
static void select_frame(void *obj, char *cmd);
static void _swap(void *obj, char *cmd);
static void zoom(void *obj, char *cmd);
static void inter(void *obj, char *cmd);
static void handle_after_write_frame(IXPServer *s, File *f);
static void handle_before_read_frame(IXPServer *s, File *f);

/* action table for /frame/?/ namespace */
Action frame_acttbl[6] = {
    {"exec", _exec},
    {"resize", __resize},
    {"select", select_frame},
    {"swap", _swap},
    {"inter", inter},
    {"zoom", zoom}
};

Frame *alloc_frame(Frame *parent, XRectangle *r, int add_frame_border, int decor)
{
    XSetWindowAttributes wa;
    char buf[MAX_BUF];
    char buf2[MAX_BUF];
    char buf3[MAX_BUF];
    static int pname = 1;
    Frame *f = (Frame *)malloc(sizeof(Frame));
    Frame *p = 0;
    XGCValues gcv;

    *f = zero_frame;
    f->id = fid++;
    f->rect = *r;
    f->parent = parent;
    f->acttbl = frame_acttbl;
    f->num_act = 6;
    f->cursor = normal_cursor;
    f->decor = decor;

    buf3[0] = '\0';
    if (parent) {
        p = root_frame(parent);
        sget_ixppath(p->files[F_PREFIX], buf3, MAX_BUF);
        snprintf(buf2, MAX_BUF, "%s/frame/%d", buf3, f->id);
        /* copy frame related info from default definition */
        snprintf(buf, MAX_BUF, "%s", buf2);
        f->files[F_PREFIX] = ixp_create(ixps, buf);
        snprintf(buf, MAX_BUF, "%s/name", buf2);
        f->files[F_NAME] = screate_ixpfile(ixps, buf, f->files[F_PREFIX]->name);
        snprintf(buf, MAX_BUF, "%s/size", buf2);
        f->files[F_SIZE] = ixp_create(ixps, buf);
        f->files[F_SIZE]->before_read = handle_before_read_frame;
        snprintf(buf, MAX_BUF, "%s/border-width", buf2);
        f->files[F_BORDER_W] =
            screate_ixpfile(ixps, buf, defaults[F_BORDER_W]->content);
        f->files[F_BORDER_W]->after_write = handle_after_write_frame;
        snprintf(buf, MAX_BUF, "%s/tab-height", buf2);
        f->files[F_TAB_H] =
            screate_ixpfile(ixps, buf, defaults[F_TAB_H]->content);
        f->files[F_TAB_H]->after_write = handle_after_write_frame;
        snprintf(buf, MAX_BUF, "%s/handle-inc", buf2);
        f->files[F_HANDLE_INC] =
            screate_ixpfile(ixps, buf, defaults[F_HANDLE_INC]->content);
        f->files[F_HANDLE_INC]->after_write = handle_after_write_frame;
        snprintf(buf, MAX_BUF, "%s/sel-style/bg-color", buf2);
        f->files[F_SELECTED_BG_COLOR] =
            screate_ixpfile(ixps, buf, defaults[F_SELECTED_BG_COLOR]->content);
        snprintf(buf, MAX_BUF, "%s/sel-style/text-align", buf2);
        f->files[F_SELECTED_TEXT_ALIGN] =
            screate_ixpfile(ixps, buf, defaults[F_SELECTED_TEXT_ALIGN]->content);
        snprintf(buf, MAX_BUF, "%s/sel-style/text-font", buf2);
        f->files[F_SELECTED_TEXT_FONT] =
            screate_ixpfile(ixps, buf, defaults[F_SELECTED_TEXT_FONT]->content);
        snprintf(buf, MAX_BUF, "%s/sel-style/text-size", buf2);
        f->files[F_SELECTED_TEXT_SIZE] =
            screate_ixpfile(ixps, buf, defaults[F_SELECTED_TEXT_SIZE]->content);
        snprintf(buf, MAX_BUF, "%s/sel-style/text-color", buf2);
        f->files[F_SELECTED_FG_COLOR] =
            screate_ixpfile(ixps, buf, defaults[F_SELECTED_FG_COLOR]->content);
        snprintf(buf, MAX_BUF, "%s/sel-style/border-color", buf2);
        f->files[F_SELECTED_BORDER_COLOR] =
            screate_ixpfile(ixps, buf, defaults[F_SELECTED_BORDER_COLOR]->content);
        snprintf(buf, MAX_BUF, "%s/norm-style/bg-color", buf2);
        f->files[F_NORMAL_BG_COLOR] =
            screate_ixpfile(ixps, buf, defaults[F_NORMAL_BG_COLOR]->content);
        snprintf(buf, MAX_BUF, "%s/norm-style/text-align", buf2);
        f->files[F_NORMAL_TEXT_ALIGN] =
            screate_ixpfile(ixps, buf, defaults[F_NORMAL_TEXT_ALIGN]->content);
        snprintf(buf, MAX_BUF, "%s/norm-style/text-font", buf2);
        f->files[F_NORMAL_TEXT_FONT] =
            screate_ixpfile(ixps, buf, defaults[F_NORMAL_TEXT_FONT]->content);
        snprintf(buf, MAX_BUF, "%s/norm-style/text-size", buf2);
        f->files[F_NORMAL_TEXT_SIZE] =
            screate_ixpfile(ixps, buf, defaults[F_NORMAL_TEXT_SIZE]->content);
        snprintf(buf, MAX_BUF, "%s/norm-style/text-color", buf2);
        f->files[F_NORMAL_FG_COLOR] =
            screate_ixpfile(ixps, buf, defaults[F_NORMAL_FG_COLOR]->content);
        snprintf(buf, MAX_BUF, "%s/norm-style/border-color", buf2);
        f->files[F_NORMAL_BORDER_COLOR] =
            screate_ixpfile(ixps, buf, defaults[F_NORMAL_BORDER_COLOR]->content);
        snprintf(buf, MAX_BUF, "%s/event/b1click", buf2);
        f->files[F_EVENT_B1CLICK] =
            screate_ixpfile(ixps, buf, defaults[F_EVENT_B1CLICK]->content);
        snprintf(buf, MAX_BUF, "%s/event/b2click", buf2);
        f->files[F_EVENT_B2CLICK] =
            screate_ixpfile(ixps, buf, defaults[F_EVENT_B2CLICK]->content);
        snprintf(buf, MAX_BUF, "%s/event/b3click", buf2);
        f->files[F_EVENT_B3CLICK] =
            screate_ixpfile(ixps, buf, defaults[F_EVENT_B3CLICK]->content);
        snprintf(buf, MAX_BUF, "%s/event/b4click", buf2);
        f->files[F_EVENT_B4CLICK] =
            screate_ixpfile(ixps, buf, defaults[F_EVENT_B4CLICK]->content);
        snprintf(buf, MAX_BUF, "%s/event/b5click", buf2);
        f->files[F_EVENT_B5CLICK] =
            screate_ixpfile(ixps, buf, defaults[F_EVENT_B5CLICK]->content);

        snprintf(buf, MAX_BUF, "%s/ignore-layout", buf2);
        if (decor) {
            f->files[F_IGNORE_LAYOUT] = screate_ixpfile(ixps, buf, "0");
            wa.override_redirect = 1;
            wa.background_pixmap = ParentRelative;
            wa.event_mask = ExposureMask | ButtonPressMask | PointerMotionMask
                | SubstructureRedirectMask | SubstructureNotifyMask;

            if (add_frame_border) {
                int bw = int_val(f->files[F_BORDER_W]->content);
                f->rect = *r;
                f->rect.width += 2 * bw;
                f->rect.height += bw + int_val(f->files[F_TAB_H]->content);
            }

            f->win = XCreateWindow(dpy, root, f->rect.x, f->rect.y,
                    f->rect.width, f->rect.height,
                    0, DefaultDepth(dpy, mon), CopyFromParent,
                    DefaultVisual(dpy, mon),
                    CWOverrideRedirect | CWBackPixmap |
                    CWEventMask, &wa);
        }
        else {
            f->files[F_IGNORE_LAYOUT] = screate_ixpfile(ixps, buf, "1");
            f->win = root;
        }
    }
    else {
        snprintf(buf2, MAX_BUF, "/page/%d", f->id);
        snprintf(buf, MAX_BUF, "%s", buf2);
        f->files[F_PREFIX] = ixp_create(ixps, buf);
        snprintf(buf, MAX_BUF, "%s/name", buf2);
        snprintf(buf3, MAX_BUF, "page %d", pname++);
        f->files[F_NAME] = screate_ixpfile(ixps, buf, buf3);
        snprintf(buf, MAX_BUF, "%s/ignore-layout", buf2);
        f->files[F_IGNORE_LAYOUT] = screate_ixpfile(ixps, buf, "0");
        f->win = root;
    }

    f->files[F_IGNORE_LAYOUT]->after_write = handle_after_write_frame;
    f->files[F_NAME]->after_write = handle_after_write_frame;
    snprintf(buf, MAX_BUF, "%s/frame", buf2);
    f->files[F_FRAME_PREFIX] = ixp_create(ixps, buf);
    snprintf(buf, MAX_BUF, "%s/frame/sel", buf2);
    f->files[F_SELECTED_FRAME] = ixp_create(ixps, buf);
    f->files[F_SELECTED_FRAME]->bind = 1;
    snprintf(buf, MAX_BUF, "%s/ctl", buf2);
    f->files[F_CTL] = ixp_create(ixps, buf);
    f->files[F_CTL]->after_write = handle_after_write;
    snprintf(buf, MAX_BUF, "%s/layout", buf2);
    f->files[F_LAYOUT] = ixp_create(ixps, buf);
    f->files[F_LAYOUT]->after_write = handle_after_write_frame;
    f->layout = alloc_layout(f->files[F_LAYOUT]->content);

    if (decor) {
        gcv.subwindow_mode = IncludeInferiors;
        gcv.function = GXxor;
        gcv.foreground = xorcolor.pixel;
        gcv.line_width = 4;
        gcv.plane_mask = AllPlanes;
        gcv.graphics_exposures = False;
        f->xorgc = XCreateGC(dpy, f->win, GCForeground | GCGraphicsExposures
                | GCFunction | GCSubwindowMode | GCLineWidth
                | GCPlaneMask, &gcv);

        install_cursor(dpy, f->win, f->cursor);
        f->gc = XCreateGC(dpy, f->win, 0, 0);
        XSync(dpy, False);
    }
    else f->gc = f->xorgc = 0;

    /* attach page */
    if (!parent) {
        /* each page references itself in focus stack */
        f->frames = r_attach(f->frames, f);
        pages = r_attach(first_page, f);
        if (!first_page) first_page = pages;
        core_files[CORE_SELECTED_PAGE]->content = f->files[F_PREFIX]->content;
        core_files[CORE_FRAME_SEL]->content = f->files[F_PREFIX]->content;
        invoke_page_event(core_files[CORE_EVENT_PAGE_CREATE]);
        invoke_page_event(core_files[CORE_EVENT_PAGE_SEL]);
    }
    else {
        parent->files[F_SELECTED_FRAME]->content = f->files[F_PREFIX]->content;
        core_files[CORE_FRAME_SEL]->content = f->files[F_PREFIX]->content;
    }
    return f;
}

void free_frame(Frame *f) {
    if (f->decor) {
        XFreeGC(dpy, f->xorgc);
        XFreeGC(dpy, f->gc);
    }
    if (zoomed == f) zoomed = 0;
    if (!f->frames) XDestroyWindow(dpy, f->win);
    else {
        /* each page contains itself in frame list once */
        f->frames = r_detach(f->frames, f);
        assert(!f->frames);
        if (first_page->item == f) {
            if (first_page->next == first_page) first_page = 0;
            else first_page = first_page->next;
        }
        pages = r_detach(pages, f);
        core_files[CORE_SELECTED_PAGE]->content = 0;
    }
    ixp_remove_file(ixps, f->files[F_PREFIX]);
    if (ixps->errstr)
        fprintf(stderr, "wmii: free_frame(): %s\n", ixps->errstr);
    if (f->layout) free_layout(f->layout);
    free(f);
}

void set_ignore_layout(Frame *f, char *flag) {
    if (f->files[F_IGNORE_LAYOUT]->content)
        free(f->files[F_IGNORE_LAYOUT]->content);
    if (!flag) return;
    f->files[F_IGNORE_LAYOUT]->content = strdup(flag);
    f->files[F_IGNORE_LAYOUT]->size = strlen(flag);
}

Cursor cursor_for_xy(Frame *f, int x, int y) {

    int n, e, w, s, tn, te, tw, ts;
    int tabh, bw;

    bw = int_val(f->files[F_BORDER_W]->content);
    tabh = int_val(f->files[F_TAB_H]->content);

    if (!f || !bw) return normal_cursor;

    /* rectangle attributes of client are used */
    w = x < bw;
    e = x >= f->rect.width - bw;
    n = y < bw;
    s = y >= f->rect.height - bw;

    tw = x < tabh;
    te = x > f->rect.width - tabh;
    tn = y < tabh;
    ts = s > f->rect.height - tabh;

    if (f->parent->frames) {
        if ((w && n) || (w && tn) || (n && tw)) return nw_cursor;
        else if ((e && n) || (e && tn) || (n && te)) return ne_cursor;
        else if ((w && s) || (w && ts) || (s && tw)) return sw_cursor;
        else if ((e && s) || (e && ts) || (s && te)) return se_cursor;
    }
#if 0 /* TODO for wmii-2 */
    else {
        Cell **c;
        Layout *l = match_layout(f->parent->layout, size_of_frame(f->parent));
        int dummy;
        Align align;
        if (w) align = WEST;
        else if (e) align = EAST;
        else if (n) align = NORTH;
        else if (s) align = SOUTH;
        else return normal_cursor;
        c = neighbors(l, f->relcell, align, &dummy);
        switch (align) {
        case WEST: w = w && c; break;
        case EAST: e = e && c; break;
        case NORTH: n = n && c; break;
        case SOUTH: s = s && c; break;
        default: break;
        }
        if (c) free(c);
    }
#endif

    if (w) return w_cursor;
    else if (e) return e_cursor;
    else if (n) return n_cursor;
    else if (s) return s_cursor;

    return normal_cursor;
}

int size_of_frame(Frame *f) {
    Ring *r;
    int size = 0;
    for (r = f->sel; r; r = r->next) {
        Frame *fr = r->item;
        if (!int_val(fr->files[F_IGNORE_LAYOUT]->content)) size++;
        if (r->next == f->sel) break;
    }
    return size;
}

static void resize_client(Frame *f, int tabh, int bw) {
    Client *c = f->client;
    int h = (f->parent && f->parent->parent) ? bw : tabh;
    c->rect.x = bw;
    c->rect.y = h;
    c->rect.width = f->rect.width - 2 * bw;
    c->rect.height = f->rect.height - bw - h;
    XMoveResizeWindow(dpy, c->win, c->rect.x, c->rect.y,
                      c->rect.width, c->rect.height);
    configure_client(c);
}

void resize(Frame *f) {

    Client *c;
    unsigned int tabh, bw;

    /* next all clients */
    tabh = f->parent ? int_val(f->files[F_TAB_H]->content) : 0;
    bw = f->parent ? int_val(f->files[F_BORDER_W]->content) : 0;

    if (f->parent && !f->decor) return; /* pages have no decor */
    if (f->parent) {
        if (f->client && f->files[F_HANDLE_INC]->content &&
            ((char *)f->files[F_HANDLE_INC]->content)[0] == '1')
        {
            c = f->client;
            /* increment stuff, see chapter 4.1.2.3 of the
             * Inter-Client Communication Conventions Manual */
            if (c->size.flags & PResizeInc) {
                int base_width = 0, base_height = 0;

                if (c->size.flags & PBaseSize) {
                    base_width = c->size.base_width;
                    base_height = c->size.base_height;
                }
                else if (c->size.flags & PMinSize) {
                    /* base_{width,height} defaults to min_{width,height} */
                    base_width = c->size.min_width;
                    base_height = c->size.min_height;
                }

                /* client_width = base_width + i * c->size.width_inc
                 * for an integer i */
                f->rect.width -=
                    (f->rect.width - 2 * bw - base_width) % c->size.width_inc;
                f->rect.height -=
                    (f->rect.height - bw - tabh - base_height) % c->size.height_inc;

            }
        }
        XMoveResizeWindow(dpy, f->win, f->rect.x,
                          f->rect.y, f->rect.width, f->rect.height);
    }
    /*fprintf(stderr, "%s", "resize()\n");*/
    if (f->client) resize_client(f, tabh, bw);
    else arrange(f, tabh, bw);
}


static void draw_tab(Frame *f, char *text, int x, int y, int w, int h, int sel) {

    Draw d = {0};

    d.screen_num = mon;
    d.drawable = f->win;
    d.gc = f->gc;
    d.rect.x = x;
    d.rect.y = y;
    d.rect.width = w;
    d.rect.height = h;
    d.text = text;
    if (sel) {
        str_to_color(&d.bg, f->files[F_SELECTED_BG_COLOR]->content);
        str_to_color(&d.fg, f->files[F_SELECTED_FG_COLOR]->content);
        str_to_color4(d.border, f->files[F_SELECTED_BORDER_COLOR]->content);
        str_to_align(&d.fnt.align, f->files[F_SELECTED_TEXT_ALIGN]->content);
        d.fnt.font = f->files[F_SELECTED_TEXT_FONT]->content;
        d.fnt.scale = (double)int_val(f->files[F_SELECTED_TEXT_SIZE]->content);
    }
    else {
        str_to_color(&d.bg, f->files[F_NORMAL_BG_COLOR]->content);
        str_to_color(&d.fg, f->files[F_NORMAL_FG_COLOR]->content);
        str_to_color4(d.border, f->files[F_NORMAL_BORDER_COLOR]->content);
        str_to_align(&d.fnt.align, f->files[F_NORMAL_TEXT_ALIGN]->content);
        d.fnt.font = f->files[F_NORMAL_TEXT_FONT]->content;
        d.fnt.scale = (double)int_val(f->files[F_NORMAL_TEXT_SIZE]->content);
    }
    draw_label(dpy, &d);
}

/* speed reasoned function for client property change */
void draw_child(Frame *f) {
    Client *c;
    Frame *fr;
    Ring *r;
    int tabh;
    int i = 0, size;
    int tw;

    if (!f->decor) return;
    c = f->client;
    fr = f->parent->parent ? f->parent : f;
    tabh = int_val(fr->files[F_TAB_H]->content);
    size = size_of_frame(fr);
    tw = fr->rect.width;
    if (size) tw /= size;
    for (r = fr->childs; r; r = r->next) {
        if (r->item == f) break;
        i++;
        if (r->next == fr->childs) return;
    }
    if (tabh) {
        draw_tab(fr, c->files[C_NAME]->content,
                 i * tw, 0, tw, tabh, f->parent->sel->item == f);
    }
}

static void draw_childs(Frame *f) {
    Ring *r;
    int tabh = f->parent ? int_val(f->files[F_TAB_H]->content) : 0;
    int i = 0, size = size_of_frame(f);
    int tw = f->rect.width;

    if (size) tw /= size;
    for (r = f->childs; r; r = r->next) {
        Frame *fr = r->item;
        if (tabh) {
            draw_tab(f, fr->client ? fr->client->files[C_NAME]->content :
                     fr->files[F_NAME]->content, i++ * tw, 0, tw, tabh,
                     f->sel == r);
        }
        draw_frame(fr);
        if (r->next == f->childs) break;
    }
    if (f->client && f->parent && !f->parent->parent && tabh) {
        draw_tab(f, f->client->files[C_NAME]->content,
                 0, 0, tw, tabh, sel == f->client);
    }
    XSync(dpy, False);
}

/**
 * Assumes following files:
 *
 * ./sel-style/text-align      "<align>"
 * ./sel-style/text-font       "<value>"
 * ./sel-style/text-size       "<int>"
 * ./sel-style/text-color      "#RRGGBBAA"
 * ./sel-style/bg-color        "#RRGGBBAA"
 * ./sel-style/border-color    "#RRGGBBAA [#RRGGBBAA [#RRGGBBAA [#RRGGBBAA]]]"
 * ./norm-style/text-align      "<align>"
 * ./norm-style/text-font       "<value>"
 * ./norm-style/text-size       "<int>"
 * ./norm-style/text-color      "#RRGGBBAA"
 * ./norm-style/bg-color        "#RRGGBBAA"
 * ./norm-style/border-color    "#RRGGBBAA [#RRGGBBAA [#RRGGBBAA [#RRGGBBAA]]]"
 */
void draw_frame(Frame *f) {

    Draw d = {0};
    int bw = f->parent ? int_val(f->files[F_BORDER_W]->content) : 0;
    XRectangle notch;

    if (!f->decor) return;
    if (bw) {
        d.screen_num = mon;
        notch.x = bw;
        notch.y = bw;
        notch.width = f->rect.width - 2 * bw;
        notch.height = f->rect.height - 2 * bw;
        d.drawable = f->win;
        d.gc = f->gc;

        /* define ground plate (i = 0) */
        if (f == top_frame(f)) {
            str_to_color(&d.bg, f->files[F_SELECTED_BG_COLOR]->content);
            str_to_color(&d.fg, f->files[F_SELECTED_FG_COLOR]->content);
            str_to_color4(d.border, f->files[F_SELECTED_BORDER_COLOR]->content);
        }
        else {
            str_to_color(&d.bg, f->files[F_NORMAL_BG_COLOR]->content);
            str_to_color(&d.fg, f->files[F_NORMAL_FG_COLOR]->content);
            str_to_color4(d.border, f->files[F_NORMAL_BORDER_COLOR]->content);
        }
        d.rect = f->rect;
        d.rect.x = d.rect.y = 0;
        d.notch = &notch;

        draw_label(dpy, &d);
    }
    draw_childs(f);
}

static void focus_nth_child(Frame *f, unsigned int n) {
    int i;
    f->sel = f->childs;
    for (i = 0; i < n; i++) f->sel = f->sel->next;
    focus(f->sel->item, 1);
}

void handle_frame_buttonpress(XButtonEvent *e, Frame *f) {

    int bindex = F_EVENT_B1CLICK - 1 + e->button;
    int size = size_of_frame(f);
    if (f->cursor == normal_cursor && e->button == Button1 && size)
        focus_nth_child(f, e->x / (f->rect.width / size));
    else focus(f, 1);
    /* frame mouse handling */
    if (f->files[bindex]->content) {
        if (core_files[CORE_ACTION_LOOKUP]->content)
            spawn_expanded(dpy, ds, core_files[CORE_ACTION_LOOKUP]->content,
                           f->files[bindex]->content);
    }
}

/* {{{ snap to border stuff */

static XRectangle *rectangles(Window rootwin, unsigned int *num)
{
    XRectangle *result = 0;
    int i, j = 0;
    Window d1, d2;
    Window *wins;
    XWindowAttributes wa;
    XRectangle r;

    if (XQueryTree(dpy, rootwin, &d1, &d2, &wins, num)) {
        result = malloc(*num * sizeof(XRectangle));
        for (i = 0; i < *num; i++) {
            if (!XGetWindowAttributes(dpy, wins[i], &wa)) continue;
            if (wa.override_redirect && (wa.map_state == IsViewable)) {
                r.x = wa.x;
                r.y = wa.y;
                r.width = wa.width;
                r.height = wa.height;
                result[j++] = r;
            }
        }
    }
    if (wins) {
        XFree(wins);
    }
    *num = j;
    return result;
}

static int check_vert_match(XRectangle *r, XRectangle *neighbor)
{
    /* check if neighbor matches edge */
    return (((neighbor->y <= r->y) && (neighbor->y + neighbor->height >= r->y))
            || ((neighbor->y >= r->y) && (r->y + r->height >= neighbor->y)));
}

static int check_horiz_match(XRectangle *r, XRectangle *neighbor)
{
    /* check if neighbor matches edge */
    return (((neighbor->x <= r->x) && (neighbor->x + neighbor->width >= r->x))
            || ((neighbor->x >= r->x) && (r->x + r->width >= neighbor->x)));
}

static void snap_move(XRectangle *r, XRectangle *rects,
                      unsigned int num, int snapw, int snaph)
{

    int i, j, w = 0, n = 0, e = 0, s = 0;

    /* snap to other windows */
    for (i = 0; i <= snapw && !(w && e); i++) {

        for (j = 0; j < num && !(w && e); j++) {

            /* check west neighbors leftwards */
            if (!w) {
                if (r->x - i == (rects[j].x + rects[j].width)) {
                    /* west edge of neighbor found, check vert match */
                    w = check_vert_match(r, &rects[j]);
                    if (w) r->x = rects[j].x + rects[j].width;
                }
            }
            /* check west neighbors rightwards */
            if (!w) {
                if (r->x + i == (rects[j].x + rects[j].width)) {
                    /* west edge of neighbor found, check vert match */
                    w = check_vert_match(r, &rects[j]);
                    if (w) r->x = rects[j].x + rects[j].width;
                }
            }

            /* check east neighbors leftwards */
            if (!e) {
                if (r->x + r->width - i == rects[j].x) {
                    /* east edge of neighbor found, check vert match */
                    e = check_vert_match(r, &rects[j]);
                    if (e) r->x = rects[j].x - r->width;
                }
            }
            /* check east neighbors rightwards */
            if (!e) {
                if (r->x + r->width + i == rects[j].x) {
                    /* east edge of neighbor found, check vert match */
                    e = check_vert_match(r, &rects[j]);
                    if (e) r->x = rects[j].x - r->width;
                }
            }

        }

        /* snap to west screen border */
        if (!w && (r->x - i == rect.x)) {
            w = 1;
            r->x = rect.x;
        }
        /* snap to west screen border */
        if (!w && (r->x + i == rect.x)) {
            w = 1;
            r->x = rect.x;
        }
        /* snap to east screen border */
        if (!e && (r->x + r->width - i == rect.width)) {
            e = 1;
            r->x = rect.x + rect.width - r->width;
        }
        if (!e && (r->x + r->width + i == rect.width)) {
            e = 1;
            r->x = rect.x + rect.width - r->width;
        }

    }

    for (i = 0; i <= snaph && !(n && s); i++) {

        for (j = 0; j < num && !(n && s); j++) {
            /* check north neighbors upwards */
            if (!n) {
                if (r->y - i == (rects[j].y + rects[j].height)) {
                    /* north edge of neighbor found, check horiz match */
                    n = check_horiz_match(r, &rects[j]);
                    if (n) r->y = rects[j].y + rects[j].height;
                }
            }
            /* check north neighbors downwards */
            if (!n) {
                if (r->y + i == (rects[j].y + rects[j].height)) {
                    /* north edge of neighbor found, check horiz match */
                    n = check_horiz_match(r, &rects[j]);
                    if (n) r->y = rects[j].y + rects[j].height;
                }
            }

            /* check south neighbors upwards */
            if (!s) {
                if (r->y + r->height - i == rects[j].y) {
                    /* south edge of neighbor found, check horiz match */
                    s = check_horiz_match(r, &rects[j]);
                    if (s) r->y = rects[j].y - r->height;
                }
            }
            /* check south neighbors downwards */
            if (!s) {
                if (r->y + r->height + i == rects[j].y) {
                    /* south edge of neighbor found, check horiz match */
                    s = check_horiz_match(r, &rects[j]);
                    if (s) r->y = rects[j].y - r->height;
                }
            }

        }

        /* snap to north screen border */
        if (!n && (r->y - i == rect.y)) {
            n = 1;
            r->y = rect.y;
        }
        if (!n && (r->y + i == rect.y)) {
            n = 1;
            r->y = rect.y;
        }
        /* snap to south screen border */
        if (!s && (r->y + r->height - i == rect.height)) {
            s = 1;
            r->y = rect.y + rect.height - r->height;
        }
        if (!s && (r->y + r->height + i == rect.height)) {
            s = 1;
            r->y = rect.y + rect.height - r->height;
        }
    }
}

/* }}} */

void move_frame(Frame *f) {

    int px = 0, py = 0, wex, wey, ex, ey, first = 1, i;
    Window dummy;
    XEvent ev;
    Frame *fp = root_frame(f);
    /* borders */
    int snapw = rect.width *
        int_val(core_files[CORE_SNAP_VALUE]->content) / 1000;
    int snaph = rect.width *
        int_val(core_files[CORE_SNAP_VALUE]->content) / 1000;
    unsigned int num;
    unsigned int dmask;
    XRectangle *rects = rectangles(root, &num);

    for (; f->parent != fp; f = f->parent);

    XQueryPointer(dpy, f->win, &dummy, &dummy, &i, &i, &wex, &wey, &dmask);
    XTranslateCoordinates(dpy, f->win, fp->win, wex, wey, &ex, &ey, &dummy);
    XFlush(dpy);
    XGrabServer(dpy);
    XGrabPointer(dpy, fp->win, False, ButtonMotionMask | ButtonReleaseMask,
                 GrabModeAsync, GrabModeAsync, None, move_cursor, CurrentTime);
    for (;;) {
        while (!XCheckMaskEvent(dpy,
                    ButtonReleaseMask | ButtonMotionMask, &ev))
        {
            usleep(20000);
            continue;
        }

        switch (ev.type) {
        case ButtonRelease:
            if (!first) {
                XDrawRectangles(dpy, fp->win, fp->xorgc, &f->rect, 1);
                resize(f);
            }
            draw_frame(pages->item);
            XSync(dpy, False);
            free(rects);
            XUngrabPointer(dpy, ev.xbutton.time);
            XUngrabServer(dpy);
            return;
            break;
        case MotionNotify:
            XTranslateCoordinates(dpy, f->win, fp->win, ev.xmotion.x,
                                  ev.xmotion.y, &px, &py, &dummy);
            if (first) first = 0;
            else {
                XDrawRectangles(dpy, fp->win, fp->xorgc, &f->rect, 1);
                XSync(dpy, False);
            }
            f->rect.x = px - ex;
            f->rect.y = py - ey;
            snap_move(&f->rect, rects, num, snapw, snaph);
            XDrawRectangles(dpy, fp->win, fp->xorgc, &f->rect, 1);
            XSync(dpy, False);
            break;
        }
    }
}

static void snap_resize(XRectangle *r, Align align, XRectangle *rects,
                        unsigned int num, int px, int py, int snapw, int snaph)
{
    int i, j, pend = 0;

    /* x */
    switch (align) {
    case NEAST:
    case EAST:
    case SEAST:
        r->width = px - r->x;
        /* snap to border */
        for (i = 0; !pend && (i < snapw); i++) {
            if (r->x + r->width - i == rect.x + rect.width) {
                r->width -= i;
                break;
            }
            if (r->x + r->width + i == rect.x + rect.width) {
                r->width += i;
                break;
            }
            for (j = 0; j < num; j++) {
                if (r->x + r->width - i == rects[j].x) {
                    pend = check_vert_match(r, &rects[j]);
                    if (pend) {
                        r->width -= i;
                        break;
                    }
                }
                if (r->x + r->width + i == rects[j].x) {
                    pend = check_vert_match(r, &rects[j]);
                    if (pend) {
                        r->width += i;
                        break;
                    }
                }
            }
        }
        break;
    case NWEST:
    case WEST:
    case SWEST:
        r->width += r->x - px;
        r->x = px;
        /* snap to border */
        for (i = 0; !pend && (i < snapw); i++) {
            if (r->x - i == rect.x) {
                r->x -= i;
                r->width += i;
                break;
            }
            if (r->x + i == rect.x) {
                r->x += i;
                r->width -= i;
                break;
            }
            for (j = 0; j < num; j++) {
                if (r->x - i == rects[j].x + rects[j].width) {
                    pend = check_vert_match(r, &rects[j]);
                    if (pend) {
                        r->x -= i;
                        r->width += i;
                        break;
                    }
                }
                if (r->x + i == rects[j].x + rects[j].width) {
                    pend = check_vert_match(r, &rects[j]);
                    if (pend) {
                        r->x += i;
                        r->width -= i;
                        break;
                    }
                }
            }
        }
        break;
    default:
        break;
    }

    /* y */
    pend = 0;
    switch (align) {
    case SWEST:
    case SOUTH:
    case SEAST:
        r->height = py - r->y;
        /* snap to border */
        for (i = 0; !pend && (i < snaph); i++) {
            if (r->y + r->height - i == rect.y + rect.height) {
                r->height -= i;
                break;
            }
            if (r->y + r->height + i == rect.y + rect.height) {
                r->height += i;
                break;
            }
            for (j = 0; j < num; j++) {
                if (r->y + r->height - i == rects[j].y) {
                    pend = check_horiz_match(r, &rects[j]);
                    if (pend) {
                        r->height -= i;
                        break;
                    }
                }
                if (r->y + r->height + i == rects[j].y) {
                    pend = check_horiz_match(r, &rects[j]);
                    if (pend) {
                        r->height += i;
                        break;
                    }
                }
            }
        }
        break;
    case NWEST:
    case NORTH:
    case NEAST:
        r->height += r->y - py;
        r->y = py;
        /* snap to border */
        for (i = 0; !pend && (i < snaph); i++) {
            if (r->y - i == rect.y) {
                r->y -= i;
                r->height += i;
                break;
            }
            if (r->y + i == rect.y) {
                r->y += i;
                r->height -= i;
                break;
            }
            for (j = 0; j < num; j++) {
                if (r->y - i == rects[j].y + rects[j].height) {
                    pend = check_horiz_match(r, &rects[j]);
                    if (pend) {
                        r->y -= i;
                        r->height += i;
                        break;
                    }
                }
                if (r->y + i == rects[j].y + rects[j].height) {
                    pend = check_horiz_match(r, &rects[j]);
                    if (pend) {
                        r->y += i;
                        r->height -= i;
                        break;
                    }
                }
            }
        }
        break;
    default:
        break;
    }
}

void _resize(Frame *f, Align align) {

    int px = 0, py = 0, pend, first = 1;
    Window dummy;
    XEvent ev;
    /* borders */
    int snapw = rect.width *
        int_val(core_files[CORE_SNAP_VALUE]->content) / 1000;
    int snaph = rect.width *
        int_val(core_files[CORE_SNAP_VALUE]->content) / 1000;
    unsigned int num;
    Frame *fp = f->parent ? f->parent : f;
    XRectangle *rects = rectangles(fp->win, &num);

    XFlush(dpy);
    XGrabServer(dpy);

    XGrabPointer(dpy, f->win, False, ButtonMotionMask | ButtonReleaseMask,
                 GrabModeAsync, GrabModeAsync, None, resize_cursor, CurrentTime);
    for (;;) {
        pend = 0;
        while (XCheckMaskEvent(dpy,
                     ButtonReleaseMask | ButtonMotionMask, &ev))
        {
            pend = 1;
        }

        if (!pend) {
            usleep(20000);
            continue;
        }

        switch (ev.type) {
        case ButtonRelease:
            if (!first) {
                XDrawRectangles(dpy, fp->win, fp->xorgc, &f->rect, 1);
                resize(f);
            }
            XUngrabPointer(dpy, ev.xbutton.time);
            XUngrabServer(dpy);
            return;
            break;
        case MotionNotify:
            XTranslateCoordinates(dpy, f->win, f->parent ? f->parent->win : root, ev.xmotion.x,
                                  ev.xmotion.y, &px, &py, &dummy);

            if (first) {
                first = 0;
            }
            else {
                XDrawRectangles(dpy, fp->win, fp->xorgc, &f->rect, 1);
                XSync(dpy, False);
            }
            snap_resize(&f->rect, align, rects, num, px, py, snapw, snaph);
            XDrawRectangles(dpy, fp->win, fp->xorgc, &f->rect, 1);
            XSync(dpy, False);
            break;
        }
    }
}

static void inter(void *obj, char *cmd) {

    Frame *f = (Frame *)obj;

    if (!f) return;
    if (!cmd) { /* fallback, use f->cursor as hint */
        if (f->cursor == normal_cursor) move_frame(f);
        else _resize(f, cursor_to_align(f->cursor));
        return;
    }

    /* maybe the user wishes following actions on different buttons */
    if (!strncmp(cmd, "resize", 7) && f->cursor != normal_cursor)
        _resize(f, cursor_to_align(f->cursor));
    else if (!strncmp(cmd, "move", 5) && f->cursor == normal_cursor)
        move_frame(f);
}

static void zoom(void *obj, char *cmd) {

    Frame *f = (Frame *)obj;
    XRectangle new;

    if (!f || !f->parent) return;
    if (zoomed == f) {
        zoomed->rect = zoom_rect;
        resize(zoomed);
        zoomed = 0;
        return;
    }
    new = f->parent->rect;
    new.x = 0;
    new.y = 0;
    zoom_rect = f->rect;
    f->rect = new;
    XRaiseWindow(dpy, f->win);
    resize(f);
    zoomed = f;
}

static void _swap(void *obj, char *cmd) {

    Frame *f = (Frame *)obj;

    if (!f->parent) return;
    f->parent->sel = r_find(f->parent->childs, f, comp_obj);
    if (f->parent->sel == f->parent->childs)
        f->parent->sel = f->parent->sel->next;
    r_swap(f->parent->childs, f->parent->sel);
    f->parent->childs = f->parent->sel;
    focus(f->parent->sel->item, 1);
    resize(f->parent);
}

void hide_page(Frame *f) {

    Ring *r;
    for (r = f->childs; r; r = r->next) {
        Frame *fr = r->item;
        if (fr->decor) XUnmapWindow(dpy, fr->win);
        else if (fr->client) XUnmapWindow(dpy, fr->client->win);
        if (r->next == f->childs) break;
    }
    XSync(dpy, False);
}

void show_page(Frame *f) {

    Ring *r;
    for (r = f->childs; r; r = r->next) {
        Frame *fr = r->item;
        if (fr->decor) XMapWindow(dpy, fr->win);
        else if (fr->client) XMapWindow(dpy, fr->client->win);
        if (r->next == f->childs) break;
    }
    XSync(dpy, False);
}

static void __resize(void *obj, char *cmd) {
    Frame *f = (Frame *)obj;
    if (f->client) invoke_class_instance_hook(f->client, "before-resize");
    if (cmd && strrchr(cmd, ',')) {
        Frame *fr = root_frame(f);
        if ((f != fr) && fr->sel) fr = fr->sel->item;
        str_to_rect(dpy, &rect, &fr->rect, cmd);
        resize(fr);
    }
    else resize(f);
    if (f->client) invoke_class_instance_hook(f->client, "after-resize");
}

static void select_frame(void *obj, char *cmd) {

    Frame *f = 0, *fr = (Frame *)obj;
    int horiz = 0;
    if (!cmd || !fr) return;

    if (!strncmp(cmd, "prev", 5)) {
        if (fr->parent) f = r_find(fr->parent->sel, fr, comp_obj)->prev->item;
        else f = r_find(pages, fr, comp_obj)->prev->item;
        horiz = 1;
    }
    else if (!strncmp(cmd, "next", 5)) { /* next */
        if (fr->parent) f = r_find(fr->parent->sel, fr, comp_obj)->next->item;
        else f = r_find(pages, fr, comp_obj)->next->item;
        horiz = 1;
    }
    else if (!strncmp(cmd, "child", 6)) {
        if (fr->sel) f = fr->sel->item;
    }
    else if (!strncmp(cmd, "parent", 7)) {
        if (fr->parent) f = fr->parent;
    }

    if (!f) return;
    if (f->parent) focus(f, 1);
    else if (horiz) focus(top_frame(f), 1);
    else focus(f, 1);
    draw_frame(pages->item);
}

static int _handle_before_read_frame(Ring *root, File *f) {

    Ring *r;

    for (r = root; r; r = r->next) {
        Frame *fr = r->item;
        if (f == fr->files[F_SIZE]) {
            char buf[64];
            snprintf(buf, 64, "%d,%d,%d,%d", fr->rect.x, fr->rect.y,
                     fr->rect.width, fr->rect.height);
            if (f->content) free(f->content);
            f->content = strdup(buf);
            f->size = strlen(buf);
            return 1;
        }
        if (fr->sel) {
            if (_handle_before_read_frame(fr->sel, f)) return 1;
        }
        if (r->next == root) break;
    }
    return 0;
}

static void handle_before_read_frame(IXPServer *s, File *f) {

    _handle_before_read_frame(pages, f);
}

static int _handle_after_write_frame(Ring *root, File *f) {

    Ring *r;

    for (r = root; r; r = r->next) {
        Frame *fr = r->item;
        if (f == fr->files[F_LAYOUT]) {
            /* layout handling begins here on changes */
            if (fr->layout) free_layout(fr->layout);
            fr->layout = alloc_layout(fr->files[F_LAYOUT]->content);
            if (apply_layout(fr) == fr) {
                focus(fr, 0);
            }
            draw_frame(pages->item);
            return 1;
        }
        else if (f == fr->files[F_TAB_H] || f == fr->files[F_BORDER_W]
                 || f == fr->files[F_HANDLE_INC])
        {
            resize(fr);
            return 1;
        }
        else if (f == fr->files[F_NAME]) {
            if (fr->frames) invoke_page_event(core_files[CORE_EVENT_PAGE_SEL]);
            else resize(fr);
            return 1;
        }
        else if (f == fr->files[F_IGNORE_LAYOUT]) {
            Frame *p = root_frame(fr);
            if (int_val(f->content) && fr->parent != p) {
                detach(fr, 0);
                detached = r_detach(detached, fr);
                attach(p, fr);
            }
        }
        if (fr->sel) {
            if (_handle_after_write_frame(fr->sel, f)) return 1;
        }
        if (r->next == root) break;
    }
    return 0;
}

static void handle_after_write_frame(IXPServer *s, File *f) {

    _handle_after_write_frame(pages, f);
}

static void attach_to_page(Frame *page, Frame *f) {
    Ring *r = 0;
    for (r = f->sel; r; r = r->next) {
        attach_to_page(page, r->item);
        if (r->next == f->sel) break;
    }
    r = r_top(page->frames, f, comp_obj);
    if (!r) page->frames = r_attach(page->frames, f);
    else page->frames = r;
}

void attach_frame_to_frame(Frame *parent, Frame *child) {

    assert(parent != child);
    /* page's frame stack */
    attach_to_page(root_frame(parent), child);
    smove_ixpfile(child->files[F_PREFIX], parent->files[F_FRAME_PREFIX]);
    /* rest */
    child->parent = parent;
    parent->sel = parent->childs = r_attach(parent->childs, child);
    if (child->rect.width > parent->rect.width ||
        child->rect.height > parent->rect.height)
        XResizeWindow(dpy, child->win, parent->rect.width, parent->rect.height);
    XReparentWindow(dpy, child->win, parent->win, child->rect.x, child->rect.y);
}

static void _detach_from_frame_stack(Frame *page, Frame *f) {
    Ring *r = 0;
    for (r = f->sel; r; r = r->next) {
        _detach_from_frame_stack(page, r->item);
        if (r->next == f->sel) break;
    }
    page->frames = r_detach(page->frames, f);
}

void detach_frame_from_frame(Frame *child) {

    Frame *parent = child->parent;
    Frame *p = root_frame(parent);
    assert(parent != child);
    _detach_from_frame_stack(p, child);
    smove_ixpfile(child->files[F_PREFIX], core_files[CORE_DETACHED]);
    XReparentWindow(dpy, child->win, root, 0, 0);
    child->relcell = 0;
    child->parent = 0;
    if (parent->sel == parent->childs)
        parent->childs = parent->sel = r_detach(parent->sel, child);
    else parent->sel = r_detach(parent->sel, child);
    if (!parent->sel) parent->childs = 0;
}

Frame *root_frame(Frame *f) {
    Frame *res = f;
    if (!f) return 0;
    while (res->parent) res =  res->parent;
    return res;
}
