/*
 * (C)opyright MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "dat.h"
#include "fns.h"

Display *dpy;
Window root;
Ring *detached;
XRectangle zoom_rect;

static Layout zero_layout = { 0 };

/*
 * Format of lstr: [(<x>,<y>,<w>,<h>)[:]]+
 */
Layout *alloc_layout(char *lstr) {
    char *p, *e;
    char **toks;
    int numt, i, j;
    Layout *res = 0, *l, *lp;

    if (!lstr) return 0;
    numt = count_toks(lstr, ':');
    if (!numt) return 0;

    toks = malloc(sizeof(char *) * numt);
    tokenize(toks, lstr, ':');

    for (i = 0; i < numt; i++) {
        l = malloc(sizeof(Layout));
        *l = zero_layout;
        p = toks[i];
        for (p = toks[i]; *p != '\0'; p++)
            if (*p == '(') l->cells++;
        l->cell = malloc(sizeof(Cell) * l->cells);
        j = 0;
        p = toks[i];
        while ((p = strchr(p, '('))) {
            p++;
            e = strchr(p, ')');
            if (!e) {
                fprintf(stderr, "wmii: cannot load unclosed layout cell '%s', ignoring\n",
                        toks[i]);
                continue;
            }
            *e = '\0';
            if (sscanf(p, "%d,%d,%d,%d", &l->cell[j].x, &l->cell[j].y,
                       &l->cell[j].w, &l->cell[j].h) != 4)
            {
                fprintf(stderr, "wmii: cannot load bogus layout cell '%s', ignoring\n", p);
                *e = ')';
                continue;
            }
            *e = ')';
            if (l->cell[j].x < 0) l->cell[j].x = 0;
            if (l->cell[j].x > 99) l->cell[j].x = 99;
            if (l->cell[j].y < 0) l->cell[j].y = 0;
            if (l->cell[j].y > 99) l->cell[j].y = 99;
            if (l->cell[j].w <= 0) l->cell[j].w = 1;
            if (l->cell[j].w > 100) l->cell[j].w = 100;
            if (l->cell[j].h <= 0) l->cell[j].h = 1;
            if (l->cell[j].h > 100) l->cell[j].h = 100;
            j++;
        }
        if (!res) res = l;
        else {
            for (lp = res; lp->next; lp = lp->next);
            if (l->cells < lp->cells) {
                fprintf(stderr, "wmii: cannot handle successor layout '%s', because it defines\n"
                        "      lesser cells than predecessor layout, ignoring\n",
                        toks[i]);
                free_layout(l);
                continue;
            }
            lp->next = l;
        }
    }
    /* cleanup */
    free_toks(toks, numt);

    return res;
}

Layout *match_layout(Layout *layout, int nclients) {

    Layout *l;
    /* find matching layout for number of clients */
    for (l = layout; l; l = l->next) {
        if (l->cells >= nclients) return l;
    }
    /* else no matching layout found, thus return 0 */
    return 0;
    /* will be used for parent propagation */
}

void free_layout(Layout *l) {
    Layout *p, *n;

    p = l;
    while (p) {
        n = p->next;
        free(p->cell);
        free(p);
        p = n;
    }
}

Frame *apply_layout(Frame *f) {

    Ring *r = 0;
    Frame *fp = 0;

    if (int_val(f->files[F_IGNORE_LAYOUT]->content)) return f;
    if (f->client) {
        /* nesting */
        Layout *l = match_layout(f->layout, size_of_frame(f));
        if (l) {
            Frame *parent = alloc_frame(f->parent, &f->rect, 0, 1);
            detach_frame_from_frame(f);
            /* swap layouts */
            swap(&f->files[F_LAYOUT]->content,
                 &parent->files[F_LAYOUT]->content);
            swap((void **)&f->files[F_LAYOUT]->size,
                 (void **)&parent->files[F_LAYOUT]->size);
            swap((void **)&f->layout, (void **)&parent->layout);
            attach_frame_to_frame(parent->parent, parent);
            attach_frame_to_frame(parent, f);
            XMapRaised(dpy, parent->win);
            resize(parent->parent);
        }
        return f;
    }

    fp = root_frame(f);
    while (f->sel) {
        Frame *fr = f->sel->item;
        if (!int_val(fr->files[F_IGNORE_LAYOUT]->content)) {
            r = r_attach(r, fr);
            detach_frame_from_frame(fr);
        }
    }

    while (r) {
        Frame *fr = r->item;
        r = r_detach(r, fr);
        attach(f, fr);
    }

    /* if parent propagation if empty frame */
    fp = f->parent;
    if (!f->childs && fp) {
        detach(f, 0);
        detached = r_detach(detached, f);
        free_frame(f);
        /* return invalidation to caller */
        return 0;
    }
    return f;
}

void arrange(Frame *f, int tabh, int bw) {

    Ring *r;
    Layout *l;
    int i = 0, w, h, ht;

    l = match_layout(f->layout, size_of_frame(f));
    ht = tabh ? tabh : bw; /* leaf frames have no tabs */
    w = f->rect.width - 2 * bw;
    h = f->rect.height - bw - ht;
    for (r = f->childs; r; r = r->next) {
        Frame *fr = r->item;
        if (!int_val(fr->files[F_IGNORE_LAYOUT]->content)) {
            if (l) {
                fr->relcell = &l->cell[i++];
                if (fr->relcell->w && fr->relcell->h) {
                    fr->rect.x = ((fr->relcell->x * w) / 100) + bw;
                    fr->rect.y = ((fr->relcell->y * h) / 100) + ht;
                    fr->rect.width = (fr->relcell->w * w) / 100;
                    fr->rect.height = (fr->relcell->h * h) / 100;
                }
            }
        }
        resize(fr);
        if (r->next == f->childs) break;
    }
}

#define CELL_THRESHOLD 2 /* 2% near a cell boundary means neighbor relationship */

/* returns all directional neighbors of cell c in layout l */
Cell **neighbors(Layout *l, Cell *o, Align align, int *res_size) {

    int i;
    Cell **result = 0;
    Cell **tmp = malloc(sizeof(Cell *) * l->cells);
    *res_size = 0;

    for (i = 0; i < l->cells; i++) {
        Cell *c = &l->cell[i];
        switch (align) {
        case WEST:
            if ((c->x + c->w < o->x) && (c->x + c->w + CELL_THRESHOLD >= o->x))
                tmp[(*res_size)++] = c;
            break;
        case EAST:
            if ((c->x > o->x + o->w) && (c->x - CELL_THRESHOLD <= o->x + o->w))
                tmp[(*res_size)++] = c;
            break;
        case NORTH:
            if ((c->y + c->h < o->y) && (c->y + c->h + CELL_THRESHOLD >= o->y))
                tmp[(*res_size)++] = c;
            break;
        case SOUTH:
            if ((c->y > o->y + o->h) && (c->y - CELL_THRESHOLD <= o->y + o->h))
                tmp[(*res_size)++] = c;
            break;
        default:
            break;
        }
    }

    result = *res_size ? malloc(sizeof(Cell *) * (*res_size)) : 0;
    for (i = 0; i < (*res_size); i++) result[i] = tmp[i];
    free(tmp);
    return result;
}
