/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>

#include "../version.h"

#include "dat.h"
#include "fns.h"

Display *dpy;
char *ds;
int mon;
int fid;
int cid;
Frame *zoomed;
Client *sel;
Window root;
XRectangle rect;
Ring *detached;
Ring *pages, *first_page; /* pages */
IXPServer *ixps;
pid_t myppid;

Atom wm_delete;
Atom wm_state;
Atom wm_change_state;
Atom wm_protocols;
Atom motif_wm_hints;
Atom net_wm_desktop;

Cursor normal_cursor;
Cursor resize_cursor;
Cursor move_cursor;
Cursor drag_cursor;
Cursor w_cursor;
Cursor e_cursor;
Cursor n_cursor;
Cursor s_cursor;
Cursor nw_cursor;
Cursor ne_cursor;
Cursor sw_cursor;
Cursor se_cursor;

File *defaults[F_LAST];
File *core_files[CORE_LAST];

char sockfile[256];

char *version[] =
{
    "wmii - window manager improved 2 - "VERSION" ("DRAW")\n"
    " (C)opyright MMIV-MMV Anselm R. Garbe\n", 0
};

static void usage()
{
    fprintf(stderr,
        "usage: wmii [-s <socket file>] [-v]\n"
        "       -s   socket file (default: %s/.wmii/ixp/wmii-%s)\n"
        "       -v   version info\n", getenv("HOME"), getenv("DISPLAY"));
    exit(1);
}

static void init_atoms() {
    wm_state = XInternAtom(dpy, "WM_STATE", False);
    wm_change_state = XInternAtom(dpy, "WM_CHANGE_STATE", False);
    wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", False);
    wm_delete = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    motif_wm_hints = XInternAtom(dpy, "_MOTIF_WM_HINTS", False);
    net_wm_desktop = XInternAtom(dpy, "_NET_WM_DESKTOP", False);
}

static void init_cursors() {
    normal_cursor = XCreateFontCursor(dpy, XC_left_ptr);
    resize_cursor = XCreateFontCursor(dpy, XC_sizing);
    move_cursor = XCreateFontCursor(dpy, XC_fleur);
    drag_cursor = XCreateFontCursor(dpy, XC_cross);
    w_cursor = XCreateFontCursor(dpy, XC_left_side);
    e_cursor = XCreateFontCursor(dpy, XC_right_side);
    n_cursor = XCreateFontCursor(dpy, XC_top_side);
    s_cursor = XCreateFontCursor(dpy, XC_bottom_side);
    nw_cursor = XCreateFontCursor(dpy, XC_top_left_corner);
    ne_cursor = XCreateFontCursor(dpy, XC_top_right_corner);
    sw_cursor = XCreateFontCursor(dpy, XC_bottom_left_corner);
    se_cursor = XCreateFontCursor(dpy, XC_bottom_right_corner);
}

static void init_defaults() {
    core_files[CORE_FRAME_SEL] = ixp_create(ixps, "/frame/sel");
    core_files[CORE_FRAME_SEL]->bind = 1;
    core_files[CORE_DETACHED] = ixp_create(ixps, "/detached/frame");
    core_files[CORE_ACTION_LOOKUP] =
        ixp_create(ixps, "/default/core/lookup/actions");
    core_files[CORE_TRANS_COLOR] = screate_ixpfile(ixps,
            "/default/core/trans-color", "#D9D9D9");
    core_files[CORE_TRANS_COLOR]->after_write = handle_after_write;
    core_files[CORE_SNAP_VALUE] = screate_ixpfile(ixps,
            "/default/core/snap-value", "20"); /* 0..1000 */
    core_files[CORE_SELECTED_PAGE] = ixp_create(ixps, "/page/sel");
    core_files[CORE_SELECTED_PAGE]->bind = 1;
    core_files[CORE_EVENT_PAGE_CREATE] = ixp_create(ixps, "/event/page-create");
    core_files[CORE_EVENT_PAGE_SEL] = ixp_create(ixps, "/event/page-select");
    core_files[CORE_EVENT_PAGE_REQ] = ixp_create(ixps, "/event/page-request");
    core_files[CORE_EVENT_PAGE_UPDATE] = ixp_create(ixps, "/event/page-update");
    defaults[F_BORDER_W] =
        screate_ixpfile(ixps, "/default/frame/border-width", DEFAULT_BORDER_W);
    defaults[F_TAB_H] =
        screate_ixpfile(ixps, "/default/frame/tab-height", DEFAULT_TAB_H);
    defaults[F_HANDLE_INC] =
        screate_ixpfile(ixps, "/default/frame/handle-inc", "0");
    defaults[F_SELECTED_BG_COLOR] = screate_ixpfile(ixps,
            "/default/frame/sel-style/bg-color", DEFAULT_F_BG);
    defaults[F_SELECTED_TEXT_ALIGN] = screate_ixpfile(ixps,
            "/default/frame/sel-style/text-align", DEFAULT_FT_ALIGN);
    defaults[F_SELECTED_TEXT_FONT] = screate_ixpfile(ixps,
            "/default/frame/sel-style/text-font", DEFAULT_FT_FAM);
    defaults[F_SELECTED_TEXT_SIZE] = screate_ixpfile(ixps,
            "/default/frame/sel-style/text-size", DEFAULT_FT_SIZE);
    defaults[F_SELECTED_FG_COLOR] = screate_ixpfile(ixps,
            "/default/frame/sel-style/text-color", DEFAULT_F_TEXT);
    defaults[F_SELECTED_BORDER_COLOR] = screate_ixpfile(ixps,
            "/default/frame/sel-style/border-color", DEFAULT_F_BORDER);
    defaults[F_NORMAL_BG_COLOR] = screate_ixpfile(ixps,
            "/default/frame/norm-style/bg-color", DEFAULT_N_BG);
    defaults[F_NORMAL_TEXT_ALIGN] = screate_ixpfile(ixps,
            "/default/frame/norm-style/text-align", DEFAULT_FT_ALIGN);
    defaults[F_NORMAL_TEXT_FONT] = screate_ixpfile(ixps,
            "/default/frame/norm-style/text-font", DEFAULT_FT_FAM);
    defaults[F_NORMAL_TEXT_SIZE] = screate_ixpfile(ixps,
            "/default/frame/norm-style/text-size", DEFAULT_FT_SIZE);
    defaults[F_NORMAL_FG_COLOR] = screate_ixpfile(ixps,
            "/default/frame/norm-style/text-color", DEFAULT_N_TEXT);
    defaults[F_NORMAL_BORDER_COLOR] = screate_ixpfile(ixps,
            "/default/frame/norm-style/border-color", DEFAULT_N_BORDER);
    defaults[F_EVENT_B1CLICK] = ixp_create(ixps, "/default/frame/event/b1click");
    defaults[F_EVENT_B2CLICK] = ixp_create(ixps, "/default/frame/event/b2click");
    defaults[F_EVENT_B3CLICK] = ixp_create(ixps, "/default/frame/event/b3click");
    defaults[F_EVENT_B4CLICK] = ixp_create(ixps, "/default/frame/event/b4click");
    defaults[F_EVENT_B5CLICK] = ixp_create(ixps, "/default/frame/event/b5click");
}

static void init_monitor() {

    mon = DefaultScreen(dpy);
    XAllocNamedColor(dpy, DefaultColormap(dpy, mon),
                     core_files[CORE_TRANS_COLOR]->content,
                     &xorcolor, &xorcolor);
    rect.x = rect.y = 0;
    rect.width = DisplayWidth(dpy, mon);
    rect.height = DisplayHeight(dpy, mon);
    root = RootWindow(dpy, mon);

    XSelectInput(dpy, root, SubstructureRedirectMask
                 | SubstructureNotifyMask | PropertyChangeMask);
    install_cursor(dpy, root, normal_cursor);
}

int handle_error(Display *dpy, XErrorEvent *e) {

    char errtxt[128];

    if ((!ixps || ixps->runlevel == HALT)
        && (e->request_code == X_ChangeWindowAttributes)
        && (e->error_code == BadAccess))
    {
        fprintf(stderr, "wmii: seems, that another wm is already running\n");
        exit(1);
    }

    XGetErrorText(dpy, e->error_code, errtxt, sizeof(errtxt));
    fprintf(stderr, "libwmii: %s (%d) opcodes %d / resource 0x%x\n",
                    errtxt, e->error_code, e->request_code,
                    (unsigned int)e->resourceid);

    if (!ixps || ixps->runlevel == HALT) {
        fprintf(stderr, "wmii: failure during initialization\n");
        exit(1);
    }
    return 0;
}

static void _cleanup(Ring *rroot) {
    Ring *r;

    for (r = rroot; r; r = r->next) {
        Frame *fr = r->item;
        if (fr->client) {
            XWindowChanges wc;
            Client *c = fr->client;
            gravitate(c, int_val(fr->files[F_TAB_H]->content),
                      int_val(fr->files[F_BORDER_W]->content), 1);
            XReparentWindow(dpy, c->win, root, c->rect.x, c->rect.y);
            wc.border_width = c->border;
            XConfigureWindow(dpy, c->win, CWBorderWidth, &wc);
        }
        if (fr->sel) _cleanup(fr->sel);
        if (fr->parent) XDestroyWindow(dpy, fr->win);
        if (r->next == rroot) break;
    }
}

static void cleanup() {

    Ring *r;
    XWindowChanges wc;

    _cleanup(pages);
    for (r = detached; r; r = r->next) { /* detached clients */
        Client *c = r->item;
        wc.border_width = c->border;
        XConfigureWindow(dpy, c->win, CWBorderWidth, &wc);
        if (r->next == detached) break;
    }
    XSetInputFocus(dpy, PointerRoot, RevertToPointerRoot, CurrentTime);
}

static void exit_cleanup() {
    if (myppid == getpid())
        unlink(sockfile);
}

static void run() {

    /* init */
    init_event_hander();

    ixps = init_server(sockfile, exit_cleanup);
    if (!ixps) {
        fatal("wmii: cannot initialize IXP server");
    }
    if (!(core_files[CORE_CTL] = ixp_create(ixps, "/ctl"))) {
        fatal("wmii: cannot connect IXP server");
    }
    core_files[CORE_CTL]->after_write = handle_after_write;

    detached = pages = first_page = 0;
    zoomed = 0;
    sel = 0;

    init_atoms();
    init_cursors();
    init_defaults();
    init_monitor();
    scan_wins();

    invoke_rc("start", spawn);
    /* main event loop */
    run_server_with_fd_support(ixps, ConnectionNumber(dpy), check_event, 0);
    cleanup();
    deinit_server(ixps);
    invoke_rc("stop", spawn_wait);
    XCloseDisplay(dpy);
}

/* }}} */

int main(int argc, char *argv[])
{
    int i;

    sockfile[0] = '\0';
    /* command line args */
    if (argc > 1) {
        for (i = 1; (i < argc) && (argv[i][0] == '-'); i++) {
            switch (argv[i][1]) {
            case 'v':
                fprintf(stdout, "%s", version[0]);
                exit(0);
                break;
            case 's':
                if (i + 1 < argc) {
                    STRLCPY(sockfile, argv[++i], sizeof(sockfile));
                }
                else {
                    usage();
                }
                break;
            default:
                usage();
                break;
            }
        }
    }

    if (!getenv("HOME")) {
        fprintf(stderr, "%s", "wmii: error: $HOME environment variable is not set\n");
        usage();
    }

    dpy = XOpenDisplay(0);
    if (!dpy) fatal("wmii: cannot open display");

    ds = display_string(dpy, DefaultScreen(dpy));
    putenv(ds);
    cid = fid = 0;
    ixps = 0;
    XSetErrorHandler(handle_error);

    setup_wmii_dir();
    if (sockfile[0] == '\0') {
        snprintf(sockfile, 256, "%s/.wmii/ixp/wmii-%s",
                 getenv("HOME"), getenv("DISPLAY"));
    }

    myppid = getpid();
    run();

    return 0;
}
