/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "../version.h"

#include "ixp.h"
#include "wmii.h"

#define     B_CTL           0
#define     B_ACTION_LOOKUP 1
#define     B_LABEL_LOOKUP  2
#define     B_SIZE          3
#define     B_LAST          4

IXPServer *ixps = 0;
Display *dpy;
char *ds;
GC gc;
Window win;
XRectangle brect;
int mon;
int displayed = 0;
pid_t myppid;
char sockfile[256];
File *files[B_LAST];

static void draw_bar();
static void new(void *obj, char *arg);
static void quit(void *obj, char *arg);
static void display(void *obj, char *arg);
static void handle_after_write(IXPServer *s, File *f);

Action acttbl[3] = {
    {"quit", quit},
    {"display", display},
    {"new", new}
};

char *version[] =
{
    "wmibar - window manager improved bar - "VERSION" ("DRAW")\n"
    "  (C)opyright MMIV-MMV Anselm R. Garbe\n", 0
};

static void usage() {
    fprintf(stderr,
        "usage: wmibar [-s <socket file>] [-v] [<x>,<y>,<width>,<height>]\n"
        "       -s     socket file (default: %s/.wmii/ixp/wmibar-%s)\n"
        "       -v     version info\n", getenv("HOME"), getenv("DISPLAY"));
    exit(1);
}

/* {{{ label functions */

/**
 * <path>/text                 "<txt value>"
 * <path>/size                 "<int>,<int>,<int>,<int>"
 * <path>/style/text-align     "<align>"
 * <path>/style/text-font      "<value>"
 * <path>/style/text-size      "<int>"
 * <path>/style/text-color     "#RRGGBBAA"
 * <path>/style/bg-color       "#RRGGBBAA"
 * <path>/style/border-color   "#RRGGBBAA #RRGGBBAA #RRGGBBAA #RRGGBBAA"
 * <path>/event/b1click        "<command>"
 * <path>/event/b2click        "<command>"
 * <path>/event/b3click        "<command>"
 * <path>/event/b4click        "<command>"
 * <path>/event/b5click        "<command>"
 * <path>/meter/display        "0|1"
 * <path>/meter/value          "<int>"
 * <path>/meter/color          "#RRGGBBAA"
 */
static void create_label(char *path) {
    File *f;
    char file[MAX_BUF];
    int i;

    snprintf(file, MAX_BUF, "%s/text", path);
    f = ixp_create(ixps, file);
    assert(f);
    f->after_write = handle_after_write;
    snprintf(file, MAX_BUF, "%s/size", path);
    screate_ixpfile(ixps, file, "0,0,1,1");
    snprintf(file, MAX_BUF, "%s/style/text-align", path);
    screate_ixpfile(ixps, file, DEFAULT_FT_ALIGN);
    snprintf(file, MAX_BUF, "%s/style/text-font", path);
    screate_ixpfile(ixps, file, DEFAULT_FT_FAM);
    snprintf(file, MAX_BUF, "%s/style/text-size", path);
    screate_ixpfile(ixps, file, DEFAULT_FT_SIZE);
    snprintf(file, MAX_BUF, "%s/style/text-color", path);
    screate_ixpfile(ixps, file, DEFAULT_N_TEXT);
    snprintf(file, MAX_BUF, "%s/style/bg-color", path);
    screate_ixpfile(ixps, file, DEFAULT_N_BG);
    snprintf(file, MAX_BUF, "%s/style/border-color", path);
    screate_ixpfile(ixps, file, DEFAULT_N_BORDER);
    for (i = 1; i < 6; i++) { /* 5 buttons events */
        snprintf(file, MAX_BUF, "%s/event/b%dclick", path, i);
        ixp_create(ixps, file);
    }
    snprintf(file, MAX_BUF, "%s/meter/display", path);
    screate_ixpfile(ixps, file, "0");
    snprintf(file, MAX_BUF, "%s/meter/value", path);
    f = screate_ixpfile(ixps, file, "0");
    f->after_write = handle_after_write;
    snprintf(file, MAX_BUF, "%s/meter/color", path);
    screate_ixpfile(ixps, file, DEFAULT_F_BG);
}

/* }}} */

/* {{{ bar actions */

static void quit(void *obj, char *arg) {
    ixps->runlevel = SHUTDOWN;
}

static void display(void *obj, char *arg) {
    if (!arg) {
        return;
    }
    displayed = int_val(arg);
    if (displayed) {
       XMapRaised(dpy, win);
       draw_bar();
    }
    else {
       XUnmapWindow(dpy, win);
       XSync(dpy, False);
    }
}

static void new(void *obj, char *arg) {
    /* fast drawing interrupt */
    create_label(arg);
}

/* }}} */

/* {{{ bar functions */

/**
 * Here we don't have fixed file references
 * (thus we walk to all necessary files), because the id
 * is userdefined..
 */
static void draw(char *path) {

    Draw d = {0};
    char buf[MAX_BUF];
    File *f;
    XRectangle r = brect;

    r.x = r.y = 0;

    if (!displayed || !files[B_LABEL_LOOKUP]->content) return;

    /* don't draw any label which is not looked up currently */
    if (strncmp((char *)files[B_LABEL_LOOKUP]->content, path,
                files[B_LABEL_LOOKUP]->size)) return;

    /* text stuff */
    snprintf(buf, MAX_BUF, "%s/text", path);
    f = ixp_walk(ixps, buf);
    if (!f) return;

    /* default stuff */
    d.text = f->content;
    d.screen_num = mon;
    d.gc = gc;
    d.drawable = win;

    /* size and misc stuff */
    snprintf(buf, MAX_BUF, "%s/size", path);
    f = ixp_walk(ixps, buf);
    str_to_rect(dpy, &r, &(d.rect), f->content);
    /* style stuff */
    snprintf(buf, MAX_BUF, "%s/style/text-align", path);
    f = ixp_walk(ixps, buf);
    str_to_align(&(d.fnt.align), f->content);
    snprintf(buf, MAX_BUF, "%s/style/text-font", path);
    f = ixp_walk(ixps, buf);
    d.fnt.font = f->content;
    snprintf(buf, MAX_BUF, "%s/style/text-size", path);
    f = ixp_walk(ixps, buf);
    d.fnt.scale = (double)int_val(f->content);
    snprintf(buf, MAX_BUF, "%s/style/text-color", path);
    f = ixp_walk(ixps, buf);
    str_to_color(&d.fg, f->content);
    snprintf(buf, MAX_BUF, "%s/style/bg-color", path);
    f = ixp_walk(ixps, buf);
    str_to_color(&d.bg, f->content);
    snprintf(buf, MAX_BUF, "%s/style/border-color", path);
    f = ixp_walk(ixps, buf);
    str_to_color4(d.border, f->content);
    snprintf(buf, MAX_BUF, "%s/meter/display", path);
    /* meter stuff */
    if ((f = ixp_walk(ixps, buf)) && (((char *)f->content)[0] == '1')) {
        Meter m;
        snprintf(buf, MAX_BUF, "%s/meter/value", path);
        if ((f = ixp_walk(ixps, buf))) {
            d.meter = &m;
            d.meter->val = int_val(f->content);
            snprintf(buf, MAX_BUF, "%s/meter/color", path);
            f = ixp_walk(ixps, buf);
            str_to_color(&d.meter->fill, f->content);
            draw_meter(dpy, &d);
        }
    }
    else draw_label(dpy, &d);
}

static void draw_bar() {

    File *label = 0;
    File *f;

    if (!displayed) return;
    if (files[B_LABEL_LOOKUP]->content)
        label = ixp_walk(ixps, files[B_LABEL_LOOKUP]->content);
    if (!label || !label->content) {
        Draw d = {0};

        /* default stuff */
        d.screen_num = mon;
        d.gc = gc;
        d.drawable = win;
        d.rect.width = brect.width;
        d.rect.height = brect.height;
        d.bg.r = d.bg.g = d.bg.b = 0.1;
        d.fg.r = d.fg.g = d.fg.b = 0.1;
        STRLCPY(d.bg.colstr, DEFAULT_F_BG, sizeof(d.bg.colstr));
        STRLCPY(d.fg.colstr, DEFAULT_F_TEXT, sizeof(d.fg.colstr));
        draw_label_noborder(dpy, &d);
    }
    else {
        char buf[512];
        for (f = label->content; f; f = f->next) {
            buf[0] = '\0';
            sget_ixppath(f, buf, sizeof(buf));
            draw(buf);
        }
    }
}

static void handle_click(XButtonPressedEvent *e) {

    File *label = 0;
    File *f, *p;
    char buf[MAX_BUF];
    char path[512];
    XRectangle r;
    XRectangle br = brect;

    if (files[B_LABEL_LOOKUP]->content)
        label = ixp_walk(ixps, files[B_LABEL_LOOKUP]->content);
    if (!label) return;

    br.x = br.y = 0;
    for (f = label->content; f; f = f->next) {
        path[0] = '\0';
        sget_ixppath(f, path, sizeof(path));
        snprintf(buf, MAX_BUF, "%s/size", path);
        if ((p = ixp_walk(ixps, buf))) {
            str_to_rect(dpy, &br, &r, p->content);

            if (is_point_in_rect(e->x, e->y, &r)) {
                /* found */
                snprintf(buf, MAX_BUF, "%s/event/b%dclick",
                         path, e->button);
                if ((p = ixp_walk(ixps, buf))) {
                    if (p->content)
                        spawn_expanded(dpy, ds,
                                       files[B_ACTION_LOOKUP]->content,
                                       p->content);
                }
            }
        }
    }
}

static void check_event(Connection *e) {

    static Atom x_rootpixmap = None;
    XEvent ev;

    while (XPending(dpy)) {
        XNextEvent(dpy, &ev);
        switch(ev.type) {
        case ButtonPress:
            handle_click(&ev.xbutton);
            break;
        case Expose:
            if (ev.xexpose.count == 0) {
                draw_bar();
            }
            break;
        case PropertyNotify:
            /* check root pixmap */
            if (x_rootpixmap == None) {
                x_rootpixmap = XInternAtom(dpy, "_XROOTPMAP_ID", True);
            }

            if ((x_rootpixmap != None)
                    && (ev.xproperty.atom == x_rootpixmap))
            {
                draw_bar();
            }
            break;
        default:
            break;
        }
    }
}

/* }}} */

/* {{{ IXP wrapper */

static void handle_after_write(IXPServer *s, File *f) {

    int i;
    size_t len;
    char buf[512];

    buf[0] = '\0';
    if (!strcmp(f->name, "text")) {
        sget_ixppath(f->parent, buf, sizeof(buf));
        draw(buf);
    }
    else if (!strcmp(f->name, "value")) {
        sget_ixppath(f->parent->parent, buf, sizeof(buf));
        draw(buf);
    }
    else if (files[B_LABEL_LOOKUP] == f) {
        draw_bar();
    }
    else if (files[B_CTL] == f) {
        for (i = 0; i < 3; i++) {
            len = strlen(acttbl[i].name);
            if (!strncmp(acttbl[i].name, (char *)f->content, len)) {
                if (strlen(f->content) > len) {
                    acttbl[i].func(0, &((char *)f->content)[len + 1]);
                }
                else {
                    acttbl[i].func(0, 0);
                }
                break;
            }
        }
    }
    check_event(0);
}

static void handle_before_read(IXPServer *s, File *f) {

    char buf[64];
    if (f != files[B_SIZE]) return;
    snprintf(buf, 64, "%d,%d,%d,%d", brect.x, brect.y, brect.width, brect.height);
    if (f->content) free(f->content);
    f->content = strdup(buf);
    f->size = strlen(buf);
}

/* }}} */

static void exit_cleanup() {
    if (myppid == getpid())
        unlink(sockfile);
}

static void run(char *size) {

    XSetWindowAttributes wa;
    XGCValues gcv;
    XRectangle rect;

    /* init */
    ixps = init_server(sockfile, exit_cleanup);
    if (!ixps) {
        fatal("wmibar: cannot initialize fs server");
    }

    if (!(files[B_CTL] = ixp_create(ixps, "/ctl"))) {
        fatal("wmibar: cannot connect IXP server");
    }
    files[B_CTL]->after_write = handle_after_write;
    files[B_ACTION_LOOKUP] = ixp_create(ixps, "/lookup/actions");
    files[B_LABEL_LOOKUP] = ixp_create(ixps, "/lookup/labels");
    files[B_LABEL_LOOKUP]->after_write = handle_after_write;
    files[B_SIZE] = ixp_create(ixps, "/size");
    files[B_SIZE]->before_read = handle_before_read;

    wa.override_redirect = 1;
    wa.background_pixmap = ParentRelative;
    wa.event_mask = ExposureMask | ButtonPressMask
                    | SubstructureRedirectMask | SubstructureNotifyMask;

    brect.x = brect.y = brect.width = brect.height = 0;
    rect.x = rect.y = 0;
    rect.width = DisplayWidth(dpy, mon);
    rect.height = DisplayHeight(dpy, mon);
    str_to_rect(dpy, &rect, &brect, size);
    if (!brect.width) {
        brect.width = DisplayWidth(dpy, mon);
    }
    if (!brect.height) {
        brect.height = 20;
    }

    win = XCreateWindow(dpy, RootWindow(dpy, mon), brect.x, brect.y,
                        brect.width, brect.height, 0, DefaultDepth(dpy, mon),
                        CopyFromParent, DefaultVisual(dpy, mon),
                        CWOverrideRedirect | CWBackPixmap |
                        CWEventMask, &wa);
    XSelectInput(dpy, DefaultRootWindow(dpy), PropertyChangeMask);
    install_cursor(dpy, win, XCreateFontCursor(dpy, XC_left_ptr));
    XSync(dpy, False);

    gcv.function = GXcopy;
    gcv.graphics_exposures = False;
    gc = XCreateGC(dpy, win, 0, 0);

    /* main event loop */
    run_server_with_fd_support(ixps, ConnectionNumber(dpy),
                               check_event, 0);
    deinit_server(ixps);
    fprintf(stderr, "%s", "wmibar: exiting\n");
    XFreeGC(dpy, gc);
    XCloseDisplay(dpy);
}

/* }}} */

int main(int argc, char * argv[]) {

    char size[64];
    int i;

    sockfile[0] = '\0';
    /* command line args */
    for (i = 1; (i < argc) && (argv[i][0] == '-'); i++) {
        switch (argv[i][1]) {
        case 'v':
            fprintf(stdout, "%s", version[0]);
            exit(0);
            break;
        case 's':
            if (i + 1 < argc) {
                STRLCPY(sockfile, argv[++i], sizeof(sockfile));
            }
            else {
                usage();
            }
            break;
        default:
            usage();
            break;
        }
    }

    if (!getenv("HOME")) {
        fprintf(stderr, "%s", "wmibar: error: $HOME environment variable is not set\n");
        usage();
    }

    dpy = XOpenDisplay(0);
    if (!dpy) fatal("wmibar: cannot open display");

    mon = DefaultScreen(dpy);
    ds = display_string(dpy, mon);
    putenv(ds);

    size[0] = '\0';
    if (argc > i) STRLCPY(size, argv[i], sizeof(size));


    setup_wmii_dir();
    if (sockfile[0] == '\0') {
        snprintf(sockfile, 256, "%s/.wmii/ixp/wmibar-%s",
                 getenv("HOME"), getenv("DISPLAY"));
    }

    myppid = getpid();
    run(size);

    return 0;
}
