/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <regex.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include "../version.h"

#include "ixp.h"
#include "wmii.h"

#define I_CTL                   0
#define I_SIZE                  1
#define I_PRE_COMMAND           2
#define I_COMMAND               3
#define I_HISTORY               4
#define I_LABEL_LOOKUP          5
#define I_ACTION_LOOKUP         6
#define I_SELECTED_BG_COLOR     7
#define I_SELECTED_TEXT_ALIGN   8
#define I_SELECTED_TEXT_FONT    9
#define I_SELECTED_TEXT_SIZE    10
#define I_SELECTED_TEXT_COLOR   11
#define I_SELECTED_BORDER_COLOR 12
#define I_NORMAL_BG_COLOR       13
#define I_NORMAL_TEXT_ALIGN     14
#define I_NORMAL_TEXT_FONT      15
#define I_NORMAL_TEXT_SIZE      16
#define I_NORMAL_TEXT_COLOR     17
#define I_NORMAL_BORDER_COLOR   18
#define I_LAST                  19

typedef struct Entry Entry;
struct Entry {
    char *label;
    char *cmd;
};

IXPServer *ixps = 0;
Display *dpy;
char *ds;
GC gc;
Window win;
XRectangle irect;
int mon;
int lid = 0;
pid_t myppid;
char sockfile[256];
File *files[I_LAST];
Ring *actions = 0;
Ring *history = 0;

static void check_event(Connection *c);
static void draw_bar();
static void handle_kpress(XKeyEvent *e);
static void set_text(char *text);
static void quit(void *obj, char *arg);
static void display(void *obj, char *arg);
static void free_entry(void *entry);
static int update_entries(char *prefix, Ring **ring);

Action acttbl[2] = {
    {"quit", quit},
    {"display", display}
};

char *version[] =
{
    "wmiinput - window manager improved input - "VERSION" ("DRAW")\n"
    " (C)opyright MMIV-MMV Anselm R. Garbe\n", 0
};

static void usage() {
    fprintf(stderr,
        "usage: wmiinput [-s <socket file>] [-r] [-v] [<x>,<y>,<width>,<height>]\n"
        "       -s       socket file (default: %s/.wmii/ixp/wmiinput-%s)\n"
        "       -v       version info\n", getenv("HOME"), getenv("DISPLAY"));
    exit(1);
}

static void add_history(char *cmd) {
    char buf[MAX_BUF];
    snprintf(buf, MAX_BUF, "/history/%ld", (long)time(0));
    history = r_attach(history, screate_ixpfile(ixps, buf, cmd));
}

/* {{{ input mode actions */

static void _exec(char *cmd) {
    char *rc = cmd;
    File *f = 0, *p = 0;

    if (!cmd || cmd[0] == '\0') return;

    add_history(cmd);
    if (files[I_LABEL_LOOKUP]->content) {
        size_t len = strlen(cmd);
        f = ixp_walk(ixps, files[I_LABEL_LOOKUP]->content);
        if (f && f->content) {
        for (p = f->content; p; p = p->next)
            if (!strncmp(p->name, cmd, len)) break;
        }
    }
    if (p && p->size) rc = cmd = p->content;
    if (files[I_PRE_COMMAND]->content) {
        size_t len = strlen(cmd) + files[I_PRE_COMMAND]->size + 2;
        rc = malloc(len);
        snprintf(rc, len, "%s %s", (char *)files[I_PRE_COMMAND]->content, cmd);
    }
    /* fallback */
    spawn_expanded(dpy, ds, files[I_ACTION_LOOKUP]->content, rc);
    /* cleanup */
    if (files[I_PRE_COMMAND]->content) free(rc);
}

static void quit(void *obj, char *arg) {
    ixps->runlevel = SHUTDOWN;
}

static void show() {
    XMapRaised(dpy, win);
    XSync(dpy, False);
    update_entries(files[I_COMMAND]->content, &actions);
    draw_bar();
    while (XGrabKeyboard(dpy, RootWindow(dpy, mon), True, GrabModeAsync,
                         GrabModeAsync, CurrentTime) != GrabSuccess)
        usleep(1000);
}

static void hide() {
    XUngrabKeyboard(dpy, CurrentTime);
    free_ring(actions, free_entry);
    actions = 0;
    set_text(0);
    XUnmapWindow(dpy, win);
    XSync(dpy, False);
}

static void display(void *obj, char *arg) {
    if (!arg) return;
    if (int_val(arg)) show();
    else hide();
    check_event(0);
}

/* }}} */

/* {{{ buffer functions */

void set_text(char *text) {

    if (files[I_COMMAND]->content) {
        free(files[I_COMMAND]->content);
        files[I_COMMAND]->content = 0;
        files[I_COMMAND]->size = 0;
    }
    if (text && strlen(text)) {
        files[I_COMMAND]->content = strdup(text);
        files[I_COMMAND]->size = strlen(text);
    }
}

static void free_entry(void *entry) {
    Entry *e = (Entry *)entry;
    if (e) {
        free(e->label);
        free(e->cmd);
        free(e);
    }
}

static int shared_prefix() {

    char c = 0;
    size_t len, max_len = 0;
    Ring *r = actions;
    int i, stop = 0;

    if (!r) return 0;
    for (; r; r = r->next) {
        len = strlen(((Entry *)r->item)->label);
        if (len > max_len) max_len = len;
        if (r->next == actions) break;
    }
    max_len++;

    for (i = 0; !stop && (i < max_len); i++) {
        for (r = actions; r && !stop; r = r->next) {
            if (r->item == actions) c = ((Entry *)r->item)->label[i];
            stop = (((Entry *)r->item)->label[i] != c)
                   && (((Entry *)r->item)->label[i] != '\0');
            c = ((Entry *)r->item)->label[i];
            if (r->next == actions) break;
        }
    }

    return i - 1;
}

static int update_entries(char *pattern, Ring **ring) {
    Entry *e = 0;
    size_t plen = pattern ? strlen(pattern) : 0;
    int matched = pattern ? plen == 0 : 1;
    int entries = 0;
    regex_t reg;
    regmatch_t match;
    File *f, *p;
    Ring *r;

    free_ring(*ring, free_entry);
    *ring = 0;

    if (!files[I_LABEL_LOOKUP]->content) return 0;
    f = ixp_walk(ixps, files[I_LABEL_LOOKUP]->content);
    if (!f || !is_directory(f)) return 0;

    if (!matched) {
       if (regcomp(&reg, pattern, REG_EXTENDED)) {
           fprintf(stderr, "wmiinput: invalid regexp pattern '%s'\n", pattern);
           return 0;
       }
    }

    /* build new actions */
    for (p = f->content; p; p = p->next) {
        if (!p->content) continue; /* ignore bogus files */
        if (matched || !regexec(&reg, p->name, 1, &match, 0)) {
            e = malloc(sizeof(Entry));
            e->label = strdup(p->name);
            e->cmd = strdup(p->content);
            r = r_attach(*ring, e);
            if (!(*ring)) *ring = r;
            entries++;
        }
    }
    if (!matched && plen) {
        for (r = *ring; r; r = r->next) {
            e = r->item;
            if (!strncmp(e->label, pattern, plen)) {
                *ring = r;
                return entries;
            }
            if (r->next == *ring) break;
        }
    }
    if (*ring) *ring = (*ring)->next;
    return entries;
}

/* creates draw structs for input mode drawing */
#define GAP     10 /* 10 px gap between actions */
static void draw_bar() {

    Draw d = {0};
    Ring *r = actions;
    int xoff, w;
    int prev_w, more_w;

    d.screen_num = mon;
    d.gc = gc;
    d.drawable = win;
    d.rect = irect;
    str_to_color(&d.bg, files[I_NORMAL_BG_COLOR]->content);
    str_to_color4(d.border, files[I_NORMAL_BORDER_COLOR]->content);
    xoff = d.rect.x + 1;
    draw_label(dpy, &d);

    /* print command */
    d.fnt.font = files[I_NORMAL_TEXT_FONT]->content;
    str_to_align(&d.fnt.align, files[I_NORMAL_TEXT_ALIGN]->content);
    d.fnt.scale = (double)int_val(files[I_NORMAL_TEXT_SIZE]->content);
    str_to_color(&d.fg, files[I_NORMAL_BG_COLOR]->content);
    str_to_color(&d.bg, files[I_NORMAL_TEXT_COLOR]->content);
    d.text = files[I_COMMAND]->content;
    w = d.text ? text_width(dpy, files[I_NORMAL_TEXT_FONT]->content,
                             int_val(files[I_NORMAL_TEXT_SIZE]->content),
                             (unsigned char *)d.text) : 0;
    d.rect.x = xoff;
    d.rect.y = 1;
    d.rect.width = w + GAP;
    d.rect.height = irect.height - 2;
    draw_label_noborder(dpy, &d);

    if (r) {
        xoff += d.rect.width + GAP;

        more_w = text_width(dpy, files[I_NORMAL_TEXT_FONT]->content,
                int_val(files[I_NORMAL_TEXT_SIZE]->content),
                (unsigned char *)">");
        prev_w = text_width(dpy, files[I_NORMAL_TEXT_FONT]->content,
                int_val(files[I_NORMAL_TEXT_SIZE]->content),
                (unsigned char *)"<");

        d.text = "<";
        d.rect.x = xoff;
        d.rect.y = 1;
        d.rect.width = prev_w + GAP;
        d.rect.height = irect.height - 2;
        draw_label_noborder(dpy, &d);
        xoff += d.rect.width;

        /* determine maximum actions */
        for (; r; r = r->next) {
            Entry *e = r->item;
            if (r == actions) {
                w = text_width(dpy, files[I_SELECTED_TEXT_FONT]->content,
                        int_val(files[I_SELECTED_TEXT_SIZE]->content),
                        (unsigned char *)e->label) + GAP;
            }
            else {
                w = text_width(dpy, files[I_NORMAL_TEXT_FONT]->content,
                        int_val(files[I_NORMAL_TEXT_SIZE]->content),
                        (unsigned char *)e->label) + GAP;
            }
            if (xoff + w > irect.width - GAP - more_w) {
                break;
            }
            d.text = e->label;
            d.rect.x = xoff;
            d.rect.y = 1;
            d.rect.width = w;
            d.rect.height = irect.height - 2;
            if (r == actions) {
                d.fnt.font = files[I_SELECTED_TEXT_FONT]->content;
                str_to_align(&d.fnt.align, files[I_SELECTED_TEXT_ALIGN]->content);
                d.fnt.scale = (double)int_val(files[I_SELECTED_TEXT_SIZE]->content);
                str_to_color(&d.bg, files[I_SELECTED_BG_COLOR]->content);
                str_to_color(&d.fg, files[I_SELECTED_TEXT_COLOR]->content);
                str_to_color4(d.border, files[I_SELECTED_BORDER_COLOR]->content);
                draw_label(dpy, &d);
            }
            else {
                d.fnt.font = files[I_NORMAL_TEXT_FONT]->content;
                str_to_align(&d.fnt.align, files[I_NORMAL_TEXT_ALIGN]->content);
                d.fnt.scale = (double)int_val(files[I_NORMAL_TEXT_SIZE]->content);
                str_to_color(&d.bg, files[I_NORMAL_BG_COLOR]->content);
                str_to_color(&d.fg, files[I_NORMAL_TEXT_COLOR]->content);
                str_to_color4(d.border, files[I_NORMAL_BORDER_COLOR]->content);
                draw_label_noborder(dpy, &d);
            }
            xoff += w;
            if (r->next == actions) break;
        }

        d.fnt.font = files[I_NORMAL_TEXT_FONT]->content;
        str_to_align(&d.fnt.align, files[I_NORMAL_TEXT_ALIGN]->content);
        d.fnt.scale = (double)int_val(files[I_NORMAL_TEXT_SIZE]->content);
        str_to_color(&d.bg, files[I_NORMAL_BG_COLOR]->content);
        str_to_color(&d.fg, files[I_NORMAL_TEXT_COLOR]->content);
        d.text = ">";
        d.rect.x = xoff;
        d.rect.y = 1;
        d.rect.width = more_w + GAP;
        d.rect.height = irect.height - 2;
        draw_label_noborder(dpy, &d);
    }
}

static void handle_kpress(XKeyEvent *e) {

    KeySym ksym;
    char buf[32];
    int num;
    char *text = files[I_COMMAND]->content;
    size_t len = text ? strlen(text) : 0;

    buf[0] = '\0';
    num = XLookupString(e, buf, 33, &ksym, 0);

    if (IsFunctionKey(ksym) || IsKeypadKey(ksym)
            || IsMiscFunctionKey(ksym) || IsPFKey(ksym)
            || IsPrivateKeypadKey(ksym)) return;

    /* first check if a control mask is omitted */
    if (e->state & ControlMask) {
        switch (ksym) {
        case XK_E:
        case XK_e:
            ksym = XK_End;
            break;
        case XK_H:
        case XK_h:
            ksym = XK_BackSpace;
            break;
        case XK_J:
        case XK_j:
            ksym = XK_Return;
            break;
        case XK_U:
        case XK_u:
            set_text(0);
            update_entries(text, &actions);
            draw_bar();
            return;
            break;
        default: /* ignore other control sequences */
            return;
            break;
        }
    }

    switch(ksym) {
    case XK_Left:
        if (!actions) return;
        actions = actions->prev;
        set_text(((Entry *)actions->item)->label);
        break;
    case XK_Tab:
        if (actions) {
            char c;
            int l = shared_prefix();
            if (l && (l > len)) {
                c = ((Entry *)actions->item)->label[l];
                ((Entry *)actions->item)->label[l] = '\0';
                set_text(((Entry *)actions->item)->label);
                ((Entry *)actions->item)->label[l] = c;
                update_entries(files[I_COMMAND]->content, &actions);
                break;
            }
        }
    case XK_Right:
        if (!actions) return;
        actions = actions->next;
        set_text(((Entry *)actions->item)->label);
        break;
    case XK_Down:
        if (history) {
            File *f;
            history = history->prev;
            f = history->item;
            set_text((char *)f->content);
        }
        update_entries(files[I_COMMAND]->content, &actions);
        break;
    case XK_Up:
        if (history) {
            File *f;
            history = history->next;
            f = history->item;
            set_text((char *)f->content);
        }
        update_entries(files[I_COMMAND]->content, &actions);
        break;
    case XK_Return:
        if (actions) _exec(((Entry *)actions->item)->label);
        else if (text) _exec(text);
    case XK_Escape:
        hide();
        break;
    case XK_BackSpace:
        if (len) {
            int entries = 0;
            Ring *r;
            size_t i = len;
            for (r = actions; r; r = r->next) {
                entries++;
                if (r->next == actions) break;
            }
            if (i) {
                do text[--i] = '\0';
                while (entries && i && entries == update_entries(text, &actions));
            }
            update_entries(text, &actions);
        }
        break;
    default:
        if ((num == 1) && !iscntrl((int)buf[0])) {
            if (!files[I_COMMAND]->size) files[I_COMMAND]->size++;
            buf[num] = '\0';
            files[I_COMMAND]->size += strlen(buf);
            files[I_COMMAND]->content = malloc(files[I_COMMAND]->size);
            ((char *)files[I_COMMAND]->content)[0] = '\0';
            if (text) STRLCAT(files[I_COMMAND]->content, text,
                              files[I_COMMAND]->size);
            STRLCAT(files[I_COMMAND]->content, buf,
                   files[I_COMMAND]->size);
            if (text) free(text);
            update_entries(files[I_COMMAND]->content, &actions);
        }
    }
    draw_bar();
}

static void check_event(Connection *c) {

    static Atom x_rootpixmap = None;
    XEvent ev;

    while (XPending(dpy)) {
        XNextEvent(dpy, &ev);
        switch(ev.type) {
        case KeyPress:
            handle_kpress(&ev.xkey);
            break;
        case Expose:
            if (ev.xexpose.count == 0) {
                draw_bar();
            }
            break;
        case PropertyNotify:
            /* check root pixmap */
            if (x_rootpixmap == None) {
                x_rootpixmap = XInternAtom(dpy, "_XROOTPMAP_ID", True);
            }

            if ((x_rootpixmap != None)
                    && (ev.xproperty.atom == x_rootpixmap))
            {
                draw_bar();
            }
            break;
        default:
            break;
        }
    }
}

/* }}} */

/* {{{ IXP callbacks */

static void handle_after_write(IXPServer *s, File *f) {
    int i;
    size_t len;

    if (files[I_CTL] == f) {
        for (i = 0; i < 2; i++) {
            len = strlen(acttbl[i].name);
            if (!strncmp(acttbl[i].name, (char *)f->content, len)) {
                if (strlen(f->content) > len) {
                    acttbl[i].func(0, &((char *)f->content)[len + 1]);
                }
                else {
                    acttbl[i].func(0, 0);
                }
                break;
            }
        }
    }
    else if (files[I_COMMAND] == f) {
        update_entries(files[I_COMMAND]->content, &actions);
        draw_bar();
    }
    check_event(0);
}

static void handle_before_read(IXPServer *s, File *f) {

    char buf[64];
    if (f != files[I_SIZE]) return;
    snprintf(buf, 64, "%d,%d,%d,%d", irect.x, irect.y, irect.width, irect.height);
    if (f->content) free(f->content);
    f->content = strdup(buf);
    f->size = strlen(buf);
}

/* }}} */

static void exit_cleanup() {
    if (myppid == getpid())
        unlink(sockfile);
}

static void run(char *size) {

    XSetWindowAttributes wa;
    XGCValues gcv;
    XRectangle rect;

    /* init */
    ixps = init_server(sockfile, exit_cleanup);
    if (!ixps) {
        fatal("wmiinput: cannot initialize fs server");
    }

    if (!(files[I_CTL] = ixp_create(ixps, "/ctl"))) {
        fatal("wmiinput: cannot connect IXP server");
    }
    files[I_CTL]->after_write = handle_after_write;
    files[I_SIZE] = ixp_create(ixps, "/size");
    files[I_SIZE]->before_read = handle_before_read;
    files[I_PRE_COMMAND] = ixp_create(ixps, "/precmd");
    files[I_COMMAND] = ixp_create(ixps, "/cmd");
    files[I_COMMAND]->after_write = handle_after_write;
    files[I_HISTORY] = ixp_create(ixps, "/history");
    add_history("");
    files[I_LABEL_LOOKUP] = ixp_create(ixps, "/lookup/labels");
    files[I_ACTION_LOOKUP] = ixp_create(ixps, "/lookup/actions");
    files[I_SELECTED_BG_COLOR] = screate_ixpfile(ixps,
        "/sel-style/bg-color", DEFAULT_F_BG);
    files[I_SELECTED_TEXT_ALIGN] =screate_ixpfile(ixps,
        "/sel-style/text-align", DEFAULT_FT_ALIGN);
    files[I_SELECTED_TEXT_FONT] = screate_ixpfile(ixps,
        "/sel-style/text-font", DEFAULT_FT_FAM);
    files[I_SELECTED_TEXT_SIZE] = screate_ixpfile(ixps,
        "/sel-style/text-size", DEFAULT_FT_SIZE);
    files[I_SELECTED_TEXT_COLOR] = screate_ixpfile(ixps,
         "/sel-style/text-color", DEFAULT_F_TEXT);
    files[I_SELECTED_BORDER_COLOR] = screate_ixpfile(ixps,
         "/sel-style/border-color", DEFAULT_F_BORDER);
    files[I_NORMAL_BG_COLOR] = screate_ixpfile(ixps,
         "/norm-style/bg-color", DEFAULT_N_BG);
    files[I_NORMAL_TEXT_ALIGN] = screate_ixpfile(ixps,
         "/norm-style/text-align", DEFAULT_FT_ALIGN);
    files[I_NORMAL_TEXT_FONT] = screate_ixpfile(ixps,
         "/norm-style/text-font", DEFAULT_FT_FAM);
    files[I_NORMAL_TEXT_SIZE] = screate_ixpfile(ixps,
         "/norm-style/text-size", DEFAULT_FT_SIZE);
    files[I_NORMAL_TEXT_COLOR] = screate_ixpfile(ixps,
         "/norm-style/text-color", DEFAULT_N_TEXT);
    files[I_NORMAL_BORDER_COLOR] = screate_ixpfile(ixps,
         "/norm-style/border-color", DEFAULT_N_BORDER);

    wa.override_redirect = 1;
    wa.background_pixmap = ParentRelative;
    wa.event_mask = ExposureMask | ButtonPressMask | KeyPressMask
                    | SubstructureRedirectMask | SubstructureNotifyMask;

    rect.x = rect.y = 0;
    rect.width = DisplayWidth(dpy, mon);
    rect.height = DisplayHeight(dpy, mon);
    str_to_rect(dpy, &rect, &irect, size);
    if (!irect.width) {
        irect.width = DisplayWidth(dpy, mon);
    }
    if (!irect.height) {
        irect.height = 20;
    }

    win = XCreateWindow(dpy, RootWindow(dpy, mon), irect.x, irect.y,
                        irect.width, irect.height, 0, DefaultDepth(dpy, mon),
                        CopyFromParent, DefaultVisual(dpy, mon),
                        CWOverrideRedirect | CWBackPixmap |
                        CWEventMask, &wa);
    XSelectInput(dpy, DefaultRootWindow(dpy), PropertyChangeMask);
    install_cursor(dpy, win, XCreateFontCursor(dpy, XC_xterm));
    XSync(dpy, False);

    irect.x = irect.y = 0;

    /* window pixmap */
    gcv.function = GXcopy;
    gcv.graphics_exposures = False;

    gc = XCreateGC(dpy, win, 0, 0);

    /* main event loop */
    run_server_with_fd_support(ixps, ConnectionNumber(dpy),
                               check_event, 0);
    deinit_server(ixps);
    XFreeGC(dpy, gc);
    XCloseDisplay(dpy);
}

/* }}} */

int main(int argc, char * argv[]) {

    char size[64];
    int i;

    sockfile[0] = '\0';
    /* command line args */
    for (i = 1; (i < argc) && (argv[i][0] == '-'); i++) {
        switch (argv[i][1]) {
        case 'v':
            fprintf(stdout, "%s", version[0]);
            exit(0);
            break;
        case 's':
            if (i + 1 < argc) {
                STRLCPY(sockfile, argv[++i], sizeof(sockfile));
            }
            else {
                usage();
            }
            break;
        default:
            usage();
            break;
        }
    }

    if (!getenv("HOME")) {
        fprintf(stderr, "%s", "wmiinput: error: $HOME environment variable is not set\n");
        usage();
    }

    dpy = XOpenDisplay(0);
    if (!dpy) fatal("wmiinput: cannot open display");

    mon = DefaultScreen(dpy);
    ds = display_string(dpy, mon);
    putenv(ds);

    size[0] = '\0';
    if (argc > i) STRLCPY(size, argv[i], sizeof(size));

    setup_wmii_dir();
    if (sockfile[0] == '\0') {
        snprintf(sockfile, 256, "%s/.wmii/ixp/wmiinput-%s",
                 getenv("HOME"), getenv("DISPLAY"));
    }

    myppid = getpid();
    run(size);

    return 0;
}
