/*
 * (C)opyright MMIV-MMV Anselm R. Garbe <garbeam at gmail dot com>
 * See LICENSE file for license details.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "../version.h"

#include "ixp.h"
#include "wmii.h"

#define R_CTL           0
#define R_BIND_LOOKUP   1
#define R_ACTION_LOOKUP 2
#define R_TEXT_ALIGN    3
#define R_TEXT_FONT     4
#define R_TEXT_SIZE     5
#define R_TEXT_COLOR    6
#define R_BG_COLOR      7
#define R_BORDER_COLOR  8
#define R_LAST          9

#define NUM_MASKS       8

#define GAP             10

#define ButtonMask      (ButtonPressMask | ButtonReleaseMask)

/* #define NO_EVILKEYS */

typedef struct Shortcut Shortcut;

struct Shortcut {
    File *file;
    unsigned long   mod;
    KeyCode         key;
    unsigned int    button;
    Shortcut        *next;
};

IXPServer *ixps = 0;
Display *dpy;
char *ds;
GC gc;
Window win;
Window root;
XRectangle brect;
XRectangle rect;
int mon;
pid_t myppid;
char sockfile[256];
Shortcut *shortcut = 0;
File *files[R_LAST];
unsigned int num_lock_mask, scroll_lock_mask, valid_mask;

static void grab_shortcut(Shortcut *s);
static void ungrab_shortcut(Shortcut *s);
static void draw_shortcut_box(char *text);
static void quit(void *obj, char *arg);

Action acttbl[1] = {
    {"quit", quit}
};

char *version[] =
{
    "wmikeys - window manager improved keys - "VERSION" ("DRAW")\n"
    "  (C)opyright MMIV-MMV Anselm R. Garbe\n", 0
};

static void center() {

    brect.x = rect.width / 2 - brect.width / 2;
    brect.y = rect.height / 2 - brect.height / 2;
}

static void usage()
{
    fprintf(stderr,
        "usage: wmikeys [-s <socket file>] [-v] [<x>,<y>,<width>,<height>]\n"
        "       -s      socket file (default: %s/.wmii/ixp/wmikeys-%s)\n"
        "       -v      version info\n", getenv("HOME"), getenv("DISPLAY"));
    exit(1);
}

/* {{{ shortcut functions */

static void init_lock_modifiers() {

    XModifierKeymap *modmap;
    KeyCode num_lock, scroll_lock;
    static int masks[NUM_MASKS] = {
        ShiftMask, LockMask, ControlMask, Mod1Mask,
        Mod2Mask, Mod3Mask, Mod4Mask, Mod5Mask
    };
    int i;

    num_lock_mask = scroll_lock_mask = 0;
    modmap = XGetModifierMapping(dpy);
    num_lock = XKeysymToKeycode(dpy, XStringToKeysym("Num_Lock"));
    scroll_lock = XKeysymToKeycode(dpy, XStringToKeysym("Scroll_Lock"));

    if (modmap && modmap->max_keypermod > 0) {
        int max = NUM_MASKS * modmap->max_keypermod;
        for (i = 0; i < max; i++) {
            if (num_lock && (modmap->modifiermap[i] == num_lock)) {
                num_lock_mask = masks[i / modmap->max_keypermod];
            }
            else if (scroll_lock && (modmap->modifiermap[i] == scroll_lock)) {
                scroll_lock_mask = masks[i / modmap->max_keypermod];
            }
        }
    }
    XFreeModifiermap(modmap);

    valid_mask = 255 & ~(num_lock_mask | scroll_lock_mask | LockMask);
}

static Shortcut *alloc_shortcut(File *f) {

    char *k;
    Shortcut *s = (Shortcut *)malloc(sizeof(Shortcut));
    s->file = f;
    k = strrchr(s->file->name, '+');
    if (k) {
        k++;
    }
    else {
        k = s->file->name;
    }
    s->key = XKeysymToKeycode(dpy, XStringToKeysym(k));
    s->mod = str_to_mod(s->file->name);
    s->button = str_to_button(s->file->name);
    s->next = 0;

    return s;
}

static void free_shortcut(Shortcut *s) {
    ungrab_shortcut(s);
    free(s);
}

/* grabs shortcut on all monitors */
static void grab_shortcut(Shortcut *s) {

    if (s->button) {
        XGrabButton(dpy, s->button, s->mod, root,
                False, ButtonMask, GrabModeAsync, GrabModeAsync,
                None, None);
#ifndef NO_EVILKEYS
        if (num_lock_mask) {
            XGrabButton(dpy, s->button, s->mod | num_lock_mask, root,
                    False, ButtonMask, GrabModeAsync, GrabModeAsync,
                    None, None);
            XGrabButton(dpy, s->button, s->mod | num_lock_mask | LockMask,
                    root, False, ButtonMask, GrabModeAsync,
                    GrabModeAsync, None, None);
        }
        if (scroll_lock_mask) {
            XGrabButton(dpy, s->button, s->mod | scroll_lock_mask, root,
                    False, ButtonMask, GrabModeAsync, GrabModeAsync,
                    None, None);
            XGrabButton(dpy, s->button, s->mod | scroll_lock_mask | LockMask,
                    root, False, ButtonMask, GrabModeAsync,
                    GrabModeAsync, None, None);
        }
        if (num_lock_mask && scroll_lock_mask) {
            XGrabButton(dpy, s->button,
                    s->mod | num_lock_mask | scroll_lock_mask, root,
                    False, ButtonMask, GrabModeAsync, GrabModeAsync,
                    None, None);
            XGrabButton(dpy, s->button,
                    s->mod | num_lock_mask | scroll_lock_mask | LockMask,
                    root, False, ButtonMask, GrabModeAsync,
                    GrabModeAsync, None, None);
        }
#endif /* NO_EVILKEYS */
    }
    else if (s->key) {
        XGrabKey(dpy, s->key, s->mod, root,
                True, GrabModeAsync, GrabModeAsync);
#ifndef NO_EVILKEYS
        if (num_lock_mask) {
            XGrabKey(dpy, s->key, s->mod | num_lock_mask, root,
                    True, GrabModeAsync, GrabModeAsync);
            XGrabKey(dpy, s->key, s->mod | num_lock_mask | LockMask, root,
                    True, GrabModeAsync, GrabModeAsync);
        }
        if (scroll_lock_mask) {
            XGrabKey(dpy, s->key, s->mod | scroll_lock_mask, root,
                    True, GrabModeAsync, GrabModeAsync);
            XGrabKey(dpy, s->key, s->mod | scroll_lock_mask | LockMask, root,
                    True, GrabModeAsync, GrabModeAsync);
        }
        if (num_lock_mask && scroll_lock_mask) {
            XGrabKey(dpy, s->key,
                    s->mod | num_lock_mask | scroll_lock_mask, root,
                    True, GrabModeAsync, GrabModeAsync);
            XGrabKey(dpy, s->key,
                    s->mod | num_lock_mask | scroll_lock_mask | LockMask,
                    root, True, GrabModeAsync, GrabModeAsync);
        }
#endif /* NO_EVILKEYS */
    }
    XSync(dpy, False);
}

/* don't handle evil keys anymore, just define more shortcuts if you
   cannot live without evil key handling */
static void ungrab_shortcut(Shortcut *s) {

    if (s->button) {
        XUngrabButton(dpy, s->button, s->mod, root);
#ifndef NO_EVILKEYS
        if (num_lock_mask) {
            XUngrabButton(dpy, s->button, s->mod | num_lock_mask, root);
            XUngrabButton(dpy, s->button, s->mod | num_lock_mask |
                    LockMask, root);
        }
        if (scroll_lock_mask) {
            XUngrabButton(dpy, s->button, s->mod | scroll_lock_mask, root);
            XUngrabButton(dpy, s->button, s->mod | scroll_lock_mask |
                    LockMask, root);
        }
        if (num_lock_mask && scroll_lock_mask) {
            XUngrabButton(dpy, s->button,
                    s->mod | num_lock_mask | scroll_lock_mask, root);
            XUngrabButton(dpy, s->button,
                    s->mod | num_lock_mask | scroll_lock_mask |
                    LockMask, root);
        }
#endif /* NO_EVILKEYS */
    }
    else if (s->key) {
        XUngrabKey(dpy, s->key, s->mod, root);
#ifndef NO_EVILKEYS
        if (num_lock_mask) {
            XUngrabKey(dpy, s->key, s->mod | num_lock_mask, root);
            XUngrabKey(dpy, s->key, s->mod | num_lock_mask |
                    LockMask, root);
        }
        if (scroll_lock_mask) {
            XUngrabKey(dpy, s->key, s->mod | scroll_lock_mask, root);
            XUngrabKey(dpy, s->key, s->mod | scroll_lock_mask |
                    LockMask, root);
        }
        if (num_lock_mask && scroll_lock_mask) {
            XUngrabKey(dpy, s->key,
                    s->mod | num_lock_mask |  scroll_lock_mask, root);
            XUngrabKey(dpy, s->key,
                    s->mod | num_lock_mask | scroll_lock_mask |
                    LockMask, root);
        }
#endif /* NO_EVILKEYS */
    }
    XSync(dpy, False);
}

static void create_shortcut(File *f) {
    Shortcut *p = shortcut;
    Shortcut *s = alloc_shortcut(f);

    if (!shortcut) {
        shortcut = s;
    }
    else {
        while (p->next) {
            p = p->next;
        }
        p->next = s;
    }
    grab_shortcut(s);
}

static void destroy_shortcut(Shortcut *s) {

    Shortcut *p = shortcut;

    if (s == shortcut) {
        shortcut = shortcut->next;
    }
    else {
        while (p->next != s) p = p->next;
        p->next = s->next;
    }
    free_shortcut(s);
}

/* }}} */

/* {{{ shortcut handling */

static void next_keystroke(unsigned long *mod, KeyCode *key) {
    XEvent e;
    KeySym sym;
    *mod = 0;
    do {
        XMaskEvent(dpy, KeyPressMask, &e);
        *mod |= e.xkey.state & valid_mask;
        *key = e.xkey.keycode;
        sym = XKeycodeToKeysym(dpy, e.xkey.keycode, 0);
    } while (IsModifierKey(sym));
}

static void emulate_key_press(unsigned long mod, KeyCode key) {

    XEvent e;
    Window client_win;
    int revert;

    XGetInputFocus(dpy, &client_win, &revert);

    e.xkey.type = KeyPress;
    e.xkey.time = CurrentTime;
    e.xkey.window = client_win;
    e.xkey.display = dpy;
    e.xkey.state = mod;
    e.xkey.keycode = key;
    XSendEvent(dpy, client_win, True, KeyPressMask, &e);
    e.xkey.type = KeyRelease;
    XSendEvent(dpy, client_win, True, KeyReleaseMask, &e);
    XSync(dpy, False);
}

static void handle_shortcut_chain(Window w, Shortcut *processed,
                                  File *first, char *prefix, int grab)
{
    unsigned long mod;
    KeyCode key;
    File *f = first;
    Shortcut *s;
    char buf[MAX_BUF];

    if (grab) {
        XMapRaised(dpy, win);
        XGrabKeyboard(dpy, w, True, GrabModeAsync,
                      GrabModeAsync, CurrentTime);
    }
    draw_shortcut_box(prefix);
    next_keystroke(&mod, &key);

    if ((processed->mod == mod) && (processed->key == key)) {
        /* double shortcut */
        emulate_key_press(mod, key);
        f = 0; /* don't proceed in next loop */
    }
    for (; f; f = f->next) {

        s = alloc_shortcut(f);
        if ((s->mod == mod) && (s->key == key)) {
            if (f->size && f->content) {
                spawn_expanded(dpy, ds, files[R_ACTION_LOOKUP]->content, f->content);
            }
            else {
                snprintf(buf, MAX_BUF, "%s/%s", prefix, f->name);
                handle_shortcut_chain(w, s, f->content, buf, 0);
            }
            break;
        }
        free_shortcut(s);
    }
    if (grab) {
        XUngrabKeyboard(dpy, CurrentTime);
        XUnmapWindow(dpy, win);
        XSync(dpy, False);
    }
}

static void handle_shortcut(Window w, unsigned long mod,
                            KeyCode key, unsigned int button)
{
    Shortcut *s;
    if (!files[R_BIND_LOOKUP]->content) return;
    for (s = shortcut; s; s = s->next) {
        if ((s->mod == mod) && (s->key == key) && (s->button == button)) {
            if (s->file->size && s->file->content) {
                spawn_expanded(dpy, ds, files[R_ACTION_LOOKUP]->content, s->file->content);
                return;
            }
            break;
        }
    }
    if (s) {
        handle_shortcut_chain(w, s, s->file->content, s->file->name, 1);
    }
}

/* }}} */

/* {{{ keys actions */

static void quit(void *obj, char *arg) {
    ixps->runlevel = SHUTDOWN;
}

static void update() {

    File *f, *p;
    Shortcut *s;
    if (!files[R_BIND_LOOKUP]->content) return;

    f = ixp_walk(ixps, files[R_BIND_LOOKUP]->content);

    if (!f || !is_directory(f)) return; /* cannot update */

    /* destroy existing shortcuts if any */
    while ((s = shortcut)) destroy_shortcut(s);

    /* create new shortcuts */
    for (p = f->content; p; p = p->next) create_shortcut(p);
}

/* }}} */

/* {{{ box functions */

/*
 * Function assumes following fs-structure:
 *
 * /box/style/text-align     "<align>"
 * /box/style/text-font      "<value>"
 * /box/style/text-size      "<int>"
 * /box/style/text-color     "#RRGGBBAA"
 * /box/style/bg-color       "#RRGGBBAA"
 */
static void draw_shortcut_box(char *text) {

    Draw d = {0};

    brect.width = text_width(dpy, files[R_TEXT_FONT]->content,
                             int_val(files[R_TEXT_SIZE]->content),
                             (unsigned char *)text) + 2 * GAP;
    center();
    XMoveResizeWindow(dpy, win, brect.x, brect.y, brect.width, brect.height);

    /* default stuff */
    d.screen_num = mon;
    d.gc = gc;
    d.drawable = win;
    str_to_align(&(d.fnt.align), files[R_TEXT_ALIGN]->content);
    d.fnt.font = (char *)files[R_TEXT_FONT]->content;
    d.fnt.scale = int_val((char *)files[R_TEXT_SIZE]->content);
    d.text = text;
    d.rect.y = 0;
    d.rect.width = brect.width;
    d.rect.height = brect.height;
    str_to_color(&d.bg, files[R_BG_COLOR]->content);
    str_to_color(&d.fg, files[R_TEXT_COLOR]->content);
    str_to_color4(d.border, files[R_BORDER_COLOR]->content);
    draw_label(dpy, &d);
}

static void check_event(Connection *c) {
    XEvent ev;

    while (XPending(dpy)) {
        XNextEvent(dpy, &ev);
        switch(ev.type) {
        case ButtonPress:
            if (ev.xbutton.state != 0) {
                handle_shortcut(ev.xbutton.root, ev.xbutton.state,
                                0, ev.xbutton.button);
            }
            break;
        case KeyPress:
            ev.xkey.state &= valid_mask;
            handle_shortcut(ev.xkey.root, ev.xkey.state,
                            ev.xkey.keycode, 0);
            break;
        default:
            break;
        }
    }
}

/* }}} */

/* {{{ IXP callbacks */

static void handle_after_write(IXPServer *s, File *f) {
    int i;
    size_t len;

    if (f == files[R_CTL]) {
        for (i = 0; i < 1; i++) {
            len = strlen(acttbl[i].name);
            if (!strncmp(acttbl[i].name, (char *)f->content, len)) {
                if (strlen(f->content) > len) {
                    acttbl[i].func(0, &((char *)f->content)[len + 1]);
                }
                else {
                    acttbl[i].func(0, 0);
                }
                break;
            }
        }
    }
    else if (f == files[R_BIND_LOOKUP]) {
        update();
    }
    check_event(0);
}

/* }}} */

static void exit_cleanup() {
    if (myppid == getpid())
        unlink(sockfile);
}

static void run(char *size) {

    XSetWindowAttributes wa;
    XGCValues gcv;

    /* init */
    ixps = init_server(sockfile, exit_cleanup);
    if (!ixps) {
        fatal("wmikeys: cannot initialize fs server");
    }

    if (!(files[R_CTL] = ixp_create(ixps, "/ctl"))) {
        fatal("wmikeys: cannot connect IXP server");
    }
    files[R_CTL]->after_write = handle_after_write;
    files[R_BIND_LOOKUP] = ixp_create(ixps, "/lookup/bindings");
    files[R_BIND_LOOKUP]->after_write = handle_after_write;
    files[R_ACTION_LOOKUP] = ixp_create(ixps, "/lookup/actions");

    files[R_TEXT_ALIGN] =
        screate_ixpfile(ixps, "/box/style/text-align", DEFAULT_FT_ALIGN);
    files[R_TEXT_FONT] =
        screate_ixpfile(ixps, "/box/style/text-font", DEFAULT_FT_FAM);
    files[R_TEXT_SIZE] =
        screate_ixpfile(ixps, "/box/style/text-size", DEFAULT_FT_SIZE);
    files[R_TEXT_COLOR] =
        screate_ixpfile(ixps, "/box/style/text-color", DEFAULT_F_TEXT);
    files[R_BG_COLOR] =
        screate_ixpfile(ixps, "/box/style/bg-color", DEFAULT_F_BG);
    files[R_BORDER_COLOR] =
        screate_ixpfile(ixps, "/box/style/border-color", DEFAULT_F_BORDER);

    wa.override_redirect = 1;
    wa.background_pixmap = ParentRelative;
    wa.event_mask = ExposureMask | ButtonPressMask
                    | SubstructureRedirectMask | SubstructureNotifyMask;

    root = RootWindow(dpy, mon);
    rect.x = rect.y = 0;
    rect.width = DisplayWidth(dpy, mon);
    rect.height = DisplayHeight(dpy, mon);
    brect.x = brect.y = -1;
    brect.width = brect.height = 0;
    str_to_rect(dpy, &rect, &brect, size);
    /* default is center position */
    if (!brect.width) {
        brect.width = 200;
    }
    if (!brect.height) {
        brect.height = 20;
    }
    center();

    init_lock_modifiers();

    win = XCreateWindow(dpy, RootWindow(dpy, mon), brect.x, brect.y,
                        brect.width, brect.height, 0, DefaultDepth(dpy, mon),
                        CopyFromParent, DefaultVisual(dpy, mon),
                        CWOverrideRedirect | CWBackPixmap |
                        CWEventMask, &wa);
    install_cursor(dpy, win, XCreateFontCursor(dpy, XC_left_ptr));
    XSync(dpy, False);

    gcv.function = GXcopy;
    gcv.graphics_exposures = False;

    gc = XCreateGC(dpy, win, 0, 0);

    /* main event loop */
    run_server_with_fd_support(ixps, ConnectionNumber(dpy),
                               check_event, 0);
    deinit_server(ixps);
    XFreeGC(dpy, gc);
    XCloseDisplay(dpy);
}

/* }}} */

int main(int argc, char * argv[]) {

    char size[64];
    int i;

    sockfile[0] = '\0';
    /* command line args */
    for (i = 1; (i < argc) && (argv[i][0] == '-'); i++) {
        switch (argv[i][1]) {
        case 'v':
            fprintf(stdout, "%s", version[0]);
            exit(0);
            break;
        case 's':
            if (i + 1 < argc) {
                STRLCPY(sockfile, argv[++i], sizeof(sockfile));
            }
            else {
                usage();
            }
            break;
        default:
            usage();
            break;
        }
    }

    if (!getenv("HOME")) {
        fprintf(stderr, "%s", "wmikeys: $HOME environment variable is not set\n");
        usage();
    }

    dpy = XOpenDisplay(0);
    if (!dpy) fatal("wmikeys: cannot open display");

    XSetErrorHandler(dummy_error_handler);
    mon = DefaultScreen(dpy);
    ds = display_string(dpy, mon);
    putenv(ds);
    size[0] = '\0';
    if (argc > i) STRLCPY(size, argv[i], sizeof(size));

    setup_wmii_dir();
    if (sockfile[0] == '\0') {
        snprintf(sockfile, 256, "%s/.wmii/ixp/wmikeys-%s",
                 getenv("HOME"), getenv("DISPLAY"));
    }

    myppid = getpid();
    run(size);

    return 0;
}
